#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - APB Drivers.
"""

from logging import getLogger
from typing import Literal

from cocotb.handle import SimHandle
from cocotb.triggers import RisingEdge
from humannum import Hex


class APBMasterDriver:
    """APB Master bus driver."""

    def __init__(
        self,
        name: str,
        clk: SimHandle,
        rst_an: SimHandle,
        paddr: SimHandle,
        pwrite: SimHandle,
        pwdata: SimHandle,
        penable: SimHandle,
        psel: SimHandle,
        prdata: SimHandle,
        pready: SimHandle,
        pslverr: SimHandle,
        log_level: int | None = None,
    ):
        self.name = name
        self.clk = clk
        self.rst_an = rst_an
        self.paddr = paddr
        self.pwrite = pwrite
        self.pwdata = pwdata
        self.prdata = prdata
        self.penable = penable
        self.psel = psel
        self.pready = pready
        self.pslverr = pslverr
        self.addr_width = len(paddr)
        self.data_width = len(pwdata)
        self.logger = getLogger(name)
        if log_level is not None:  # important explicit check for None as 0 would be a valid value
            self.logger.setLevel(log_level)

    async def write(
        self,
        addr: int,
        data: int,
    ) -> bool:
        """APB Write."""
        self.paddr.value = addr
        self.pwdata.value = data
        self.pwrite.value = 1
        self.penable.value = 1
        self.psel.value = 0
        await RisingEdge(self.clk)
        self.psel.value = 1
        await RisingEdge(self.clk)
        while self.pready == 0:
            await RisingEdge(self.clk)
        self.paddr.value = 0
        self.pwdata.value = 0xDEADDEAD
        self.pwrite.value = 0
        self.penable.value = 0
        self.psel.value = 0
        err_resp = bool(self.pslverr.value)
        err = " ERROR" if err_resp else ""
        hexaddr = str(Hex(addr, self.addr_width))
        hexdata = str(Hex(data, self.data_width))
        self.logger.info(f"=MST WRITE{err}= address: {hexaddr} data: {hexdata}")
        return err_resp

    async def read(self, addr: int) -> tuple[bool, int]:
        """APB Read."""
        self.paddr.value = addr
        self.pwrite.value = 0
        await RisingEdge(self.clk)
        self.penable.value = 1
        await RisingEdge(self.clk)
        self.psel.value = 1
        await RisingEdge(self.clk)
        while self.pready == 0:
            await RisingEdge(self.clk)
        self.penable.value = 0
        self.psel.value = 0
        rdata = self.prdata.value.integer
        err_resp = bool(self.pslverr.value)
        err = " ERROR" if err_resp else ""
        hexaddr = str(Hex(addr, self.addr_width))
        hexdata = str(Hex(rdata, self.data_width))
        self.logger.info(f"=MST READ{err}= address: {hexaddr} data: {hexdata}")
        return (err_resp, rdata)

    async def reset(self):
        """Reset APB Master."""
        self.paddr.value = 0
        self.pwrite.value = 0
        self.pwdata.value = 0


class APBSlaveDriver:
    """Active APB Slave that can respond to Master requests."""

    def __init__(
        self,
        name: str,
        clk: SimHandle,
        rst_an: SimHandle,
        paddr: SimHandle,
        pwrite: SimHandle,
        pwdata: SimHandle,
        penable: SimHandle,
        psel: SimHandle,
        prdata: SimHandle,
        pready: SimHandle,
        pslverr: SimHandle,
        pstrb: SimHandle = None,
        pready_delay: int = 0,
        size_bytes: int = 1024,
        err_addr: dict[Literal["r", "w", "rw"], list[int]] | None = None,
        log_level: int | None = None,
    ):
        """APB Slave Init."""
        self.name = name
        self.clk = clk
        self.rst_an = rst_an
        self.paddr = paddr
        self.pwrite = pwrite
        self.pwdata = pwdata
        self.prdata = prdata
        self.penable = penable
        self.psel = psel
        self.pready = pready
        self.pslverr = pslverr
        self.pstrb = pstrb
        self.err_addr = err_addr
        self.addr_width = len(paddr)
        self.data_width = len(pwdata)
        self.byte_width = len(pwdata) // 8

        self.logger = getLogger(name)
        if log_level is not None:  # important explicit check for None as 0 would be a valid value
            self.logger.setLevel(log_level)

        self.mem = bytearray(size_bytes)  # Initialize memory
        self.pready_delay = pready_delay  # Delay for PREADY signal to simulate longer access times
        self.state = 0

    def _read(self, addr: int) -> int:
        """APB Read."""
        bwidth = self.byte_width
        addr = addr & ((1 << self.addr_width) - bwidth)  # align address to data width
        bytes = self.mem[addr : addr + bwidth]
        rdata = int.from_bytes(bytes, "little")
        hexaddr = str(Hex(addr, self.addr_width))
        hexdata = str(Hex(rdata, self.data_width))
        hexbytes = ",".join(str(Hex(x, 8)) for x in bytes)
        self.logger.info(f"=SLV READ= address: {hexaddr} data: {hexdata} data (bytes): {hexbytes}")
        return rdata

    def _write(self, addr: int, data: int, strb: int) -> None:
        """APB Write."""
        bwidth = self.byte_width
        addr = addr & ((1 << self.addr_width) - bwidth)  # align address to data width
        strbs = [(strb >> i) & 1 for i in range(bwidth)]
        bytes = bytearray(int.to_bytes(data, self.byte_width, "little"))
        for b in range(bwidth):
            if not strbs[b]:
                bytes[b] = self.mem[addr + b]  # replace with read value
        hexaddr = str(Hex(addr, self.addr_width))
        hexdata = str(Hex(data, self.data_width))
        hexbytes = ",".join(str(Hex(x, 8)) for x in bytes)
        self.logger.info(f"=SLV WRITE= address: {hexaddr} data: {hexdata} data (bytes): {hexbytes} strb: {strbs}")
        self.mem[addr : addr + bwidth] = bytes

    def _check_err_addr(self, paddr: int, pwrite: int) -> bool:
        """Check for Error Address."""
        if self.err_addr is None:
            return False
        if pwrite:
            if (paddr in self.err_addr.get("w", [])) or (paddr in self.err_addr.get("rw", [])):
                return True
        elif (paddr in self.err_addr.get("r", [])) or (paddr in self.err_addr.get("rw", [])):
            return True
        return False

    async def run(self):
        """Slave Main Loop."""
        self.pready.value = 1
        self.prdata.value = 0xDEADDEAD
        self.pslverr.value = 0
        allstrb = (1 << self.byte_width) - 1
        while True:
            await RisingEdge(self.clk)
            if self.state:
                for _ in range(self.pready_delay):  # delay the answer if configured
                    self.pready.value = 0
                    await RisingEdge(self.clk)
                self.pready.value = 1
                self.pslverr.value = 0
                if (self.state == 1) and self.pwrite.value:
                    self._write(
                        self.paddr.value.integer,
                        self.pwdata.value.integer,
                        self.pstrb.value.integer if self.pstrb else allstrb,
                    )
            # Check if there's an APB request starting
            if self.psel.value and not self.penable.value:
                self.state = 1
            else:
                self.state = 0
            if self.state and self._check_err_addr(self.paddr.value, self.pwrite.value):
                acc = "WRITE" if self.pwrite.value else "READ"
                hexaddr = str(Hex(self.paddr.value.integer, self.addr_width))
                self.logger.info(f"=SLV ERROR RESP for {acc}= address: {hexaddr}")
                self.pslverr.value = 1
                self.state = 2

            if (self.state == 1) and not self.pwrite.value:
                # Handle read request (need to apply read value for next cycle)
                rdata = self._read(self.paddr.value.integer)
                self.prdata.value = rdata
            else:
                self.prdata.value = 0xDEADBEEF

    def set_pready_delay(self, delay):
        """Set pready Delay."""
        self.pready_delay = delay

    def set_data(self, data):
        """Preload Slave Memory."""
        self.mem = data

    def log_data(self, start_addr=0, end_addr=None, chunk_size=16):
        """Request logging of Slave Memory content."""
        if end_addr is None:
            end_addr = len(self.mem) - 1
        elif end_addr > len(self.mem) - 1:
            self.logger.error(
                f"=MEMORY CONTENTS= Provided end_addr {hex(end_addr)} to log_data is beyond size of slave memory."
            )
            return
        for i in range(start_addr, end_addr + 1, chunk_size):
            hexstart = str(Hex(i, self.addr_width))
            hexend = str(Hex(i + chunk_size - 1, self.addr_width))
            memcts = ",".join(str(Hex(x, 8)) for x in self.mem[i : min(i + chunk_size, end_addr + 1)])
            self.logger.info(f"=MEMORY CONTENTS= {hexstart}-{hexend} [{memcts}]")



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - AHB Drivers.
"""

from collections.abc import Iterable
from enum import IntEnum
from logging import getLogger
from typing import Literal

from cocotb.handle import SimHandle
from cocotb.triggers import RisingEdge
from humannum import Hex


class BurstType(IntEnum):
    """HBURST encoding as per AHB spec."""

    SINGLE = 0  # Single transfer
    INCR = 1  # Unknown length burst
    WRAP4 = 2  # Four-beat wrap burst
    INCR4 = 3  # Four-beat incrementing burst
    WRAP8 = 4  # Eight-beat wrap burst
    INCR8 = 5  # Eight-beat incrementing burst
    WRAP16 = 6  # Sixteen-beat wrap burst
    INCR16 = 7  # Sixteen-beat incrementing burst


class TransType(IntEnum):
    """HTRANS encoding as per AHB spec."""

    IDLE = 0  # No transfer
    BUSY = 1  # Busy transfer
    NONSEQ = 2  # Non-sequential transfer
    SEQ = 3  # Sequential transfer


class SizeType(IntEnum):
    """HSIZE encoding as per AHB spec."""

    BYTE = 0  # 8-bit
    HALFWORD = 1  # 16-bit
    WORD = 2  # 32-bit
    DOUBLEWORD = 3  # 64-bit
    WORD4 = 4  # 128-bit
    WORD8 = 5  # 256-bit
    WORD16 = 6  # 512-bit
    WORD32 = 7  # 1024-bit


def _prep_addr_iter(addr: int, burst_length: int, size: SizeType, burst_type=BurstType) -> tuple[int, int, int, int]:
    """Prepare Address Iterations."""
    match burst_type:
        case BurstType.INCR16 | BurstType.WRAP16:
            mask = (1 << (size + 4)) - 1
            len = 16
        case BurstType.INCR8 | BurstType.WRAP8:
            mask = (1 << (size + 3)) - 1
            len = 8
        case BurstType.INCR4 | BurstType.WRAP4:
            mask = (1 << (size + 2)) - 1
            len = 4
        case BurstType.INCR:
            mask = -1
            len = burst_length
        case BurstType.SINGLE:
            mask = -1
            len = 1
    base = addr & ~mask
    offs = addr & mask
    return (base, offs, mask, len)


def _check_bus_acc(data_width: int, addr: int, offs: int, size: SizeType, burst_type: BurstType) -> None:
    """Check AHB Bus Access."""
    assert data_width >= (8 << size), f"Size argument {size!r} -> {8<<size} too big for data width of {data_width}!"
    assert (addr & ((1 << size) - 1)) == 0, f"Address {addr:x} is not aligned to size argument {size!r}!"
    assert (burst_type not in (BurstType.INCR16, BurstType.INCR8, BurstType.INCR4)) or (
        offs == 0
    ), f"Address {addr:x} is not aligned to BurstType {burst_type!r} at size {size!r}!"


def _bottomless(value: int) -> Iterable[int]:
    """Bottomless Iterator."""
    while True:
        yield value


class SlaveFsmState(IntEnum):
    """Internal AHB Slave State."""

    IDLE = 0
    ACTIVE = 1


class AHBMasterDriver:
    """AHB Master bus driver."""

    def __init__(
        self,
        name: str,
        clk: SimHandle,
        rst_an: SimHandle,
        haddr: SimHandle,
        hwrite: SimHandle,
        hwdata: SimHandle,
        htrans: SimHandle,
        hburst: SimHandle,
        hsize: SimHandle,
        hrdata: SimHandle,
        hready: SimHandle,
        hresp: SimHandle,
        hsel: SimHandle = None,
        hprot: SimHandle = None,
        hwstrb: SimHandle = None,
        hreadyout: SimHandle = None,
        hnonsec: SimHandle = None,
        hmastlock: SimHandle = None,
        hexcl: SimHandle = None,
        hmaster: SimHandle = None,
        hexokay: SimHandle = None,
        hauser: SimHandle = None,
        log_level: int | None = None,
    ):
        self.name = name
        self.clk = clk
        self.rst_an = rst_an
        self.haddr = haddr
        self.hwrite = hwrite
        self.hwdata = hwdata
        self.htrans = htrans
        self.hburst = hburst
        self.hsize = hsize
        self.hrdata = hrdata
        self.hready = hready
        self.hresp = hresp
        # optional signals:
        self.hsel = hsel
        self.hprot = hprot
        self.hwstrb = hwstrb
        self.hreadyout = hreadyout
        self.hnonsec = hnonsec
        self.hmastlock = hmastlock
        self.hexcl = hexcl
        self.hmaster = hmaster
        self.hexokay = hexokay
        self.hauser = hauser
        self.addr_width = len(haddr)
        self.data_width = len(hwdata)

        self.logger = getLogger(name)
        if log_level is not None:  # important explicit check for None as 0 would be a valid value
            self.logger.setLevel(log_level)

    async def write(  # noqa: C901, PLR0912
        self,
        addr: int,
        data: int | Iterable,
        wstrb: int | Iterable | None = None,
        size: SizeType = SizeType.WORD,
        burst_length: int = 1,
        burst_type: BurstType = BurstType.SINGLE,
        hauser: int = 0,
        hmaster: int = 5,
    ) -> bool:
        """AHB Write (Burst)."""
        base, offs, mask, burst_length = _prep_addr_iter(
            addr=addr, burst_length=burst_length, size=size, burst_type=burst_type
        )
        _check_bus_acc(data_width=self.data_width, addr=addr, offs=offs, size=size, burst_type=burst_type)

        err_resp = False
        if isinstance(data, int):
            log_data = [data]
            data = iter((data,))
        else:
            log_data = list(data)
            data = iter(data)
        allstrb = (1 << (1 << size)) - 1
        if wstrb is None:
            wstrb = iter(_bottomless(allstrb))
        elif isinstance(wstrb, int):
            wstrb = iter(_bottomless(wstrb))
        else:
            wstrb = iter(wstrb)
        shift_mask = self.data_width - 1
        self.haddr.value = base + offs
        if self.hauser:
            self.hauser.value = hauser
        if self.hmaster:
            self.hmaster.value = hmaster
        self.hwdata.value = 0xDEADDEAD
        if self.hwstrb:
            self.hwstrb.value = 0
        self.hwrite.value = 1
        if self.hsel:
            self.hsel.value = 1
        if self.hreadyout:
            self.hreadyout.value = 1
        self.htrans.value = TransType.NONSEQ
        self.hburst.value = burst_type
        self.hsize.value = size
        await RisingEdge(self.clk)
        if self.hreadyout:
            self.hreadyout = self.hready
        for _ in range(burst_length - 1):
            self.htrans.value = TransType.SEQ
            self.hwdata.value = next(data) << ((offs << 3) & shift_mask)
            if self.hwstrb:
                self.hwstrb.value = next(wstrb) << (offs & (shift_mask >> 3))
            offs = (offs + (1 << size)) & mask
            self.haddr.value = base + offs
            if self.hauser:
                self.hauser.value = hauser
            if self.hmaster:
                self.hmaster.value = hmaster
            await RisingEdge(self.clk)
            while self.hready == 0:
                await RisingEdge(self.clk)
            if self.hresp.value:
                err_resp = True
        if self.hsel:
            self.hsel.value = 0
        self.haddr.value = 0
        if self.hmaster:
            self.hmaster.value = 0
        self.hwdata.value = next(data) << ((offs << 3) & shift_mask)
        if self.hwstrb:
            self.hwstrb.value = next(wstrb) << (offs & (shift_mask >> 3))
        self.hwrite.value = 0
        self.htrans.value = TransType.IDLE
        self.hburst.value = BurstType.SINGLE
        self.hsize.value = SizeType.BYTE
        await RisingEdge(self.clk)
        if self.hreadyout:
            self.hreadyout = self.hready
        while self.hready == 0:
            await RisingEdge(self.clk)
        if self.hresp.value:
            err_resp = True
        self.hwdata.value = 0xDEADDEAD
        if self.hwstrb:
            self.hwstrb.value = 0
        self._info_log(
            rw="WRITE",
            addr=addr,
            burst_type=burst_type,
            burst_length=burst_length,
            size=size,
            data=log_data,
            err_resp=err_resp,
        )
        return err_resp

    async def read(  # noqa: C901
        self,
        addr: int,
        burst_length: int = 1,
        size: SizeType = SizeType.WORD,
        burst_type: BurstType = BurstType.SINGLE,
        hauser: int = 0,
        hmaster: int = 0,
    ) -> tuple[bool, tuple[int]]:
        """AHB Read (Burst)."""
        base, offs, mask, burst_length = _prep_addr_iter(
            addr=addr, burst_length=burst_length, size=size, burst_type=burst_type
        )
        _check_bus_acc(data_width=self.data_width, addr=addr, offs=offs, size=size, burst_type=burst_type)

        rdata = []
        err_resp = False
        poffs = offs
        shift_mask = self.data_width - 1
        szmsk = (1 << (8 << size)) - 1
        self.haddr.value = base + offs
        if self.hauser:
            self.hauser.value = hauser
        if self.hmaster:
            self.hmaster.value = hmaster
        if self.hsel:
            self.hsel.value = 1
        self.hwrite.value = 0
        self.htrans.value = TransType.NONSEQ
        self.hburst.value = burst_type
        self.hsize.value = size
        await RisingEdge(self.clk)

        for _ in range(burst_length - 1):
            self.htrans.value = TransType.SEQ
            offs = (offs + (1 << size)) & mask
            self.haddr.value = base + offs
            if self.hauser:
                self.hauser.value = hauser
            if self.hmaster:
                self.hmaster.value = hmaster
            await RisingEdge(self.clk)
            while self.hready == 0:
                await RisingEdge(self.clk)
            rdata.append((self.hrdata.value.integer >> ((poffs << 3) & shift_mask)) & szmsk)
            if self.hresp.value:
                err_resp = True
            poffs = offs
        if self.hsel:
            self.hsel.value = 0
        self.haddr.value = 0
        self.htrans.value = TransType.IDLE
        self.hsize.value = SizeType.BYTE
        await RisingEdge(self.clk)
        while self.hready == 0:
            await RisingEdge(self.clk)
        rdata.append((self.hrdata.value.integer >> ((poffs << 3) & shift_mask)) & szmsk)
        if self.hresp.value:
            err_resp = True
        self._info_log(
            rw="READ",
            addr=addr,
            burst_type=burst_type,
            burst_length=burst_length,
            size=size,
            data=rdata,
            err_resp=err_resp,
        )
        return (err_resp, tuple(rdata))

    async def reset(self):
        """Reset AHB Master."""
        if self.hsel:
            self.hsel.value = 0
        self.hwrite.value = 0
        self.hwdata.value = 0
        self.htrans.value = 0  # IDLE
        self.hburst.value = 0
        self.hprot.value = 0

    def calc_wrmem(
        self,
        offs: int,
        size: SizeType,
        blen: int,
        mmask: int,
        wdata: int | Iterable[int],
        wstrb: int | Iterable[int] | None = None,
        mem: bytearray | None = None,
    ) -> bytearray:
        """Calculate Reference Write Data for a Burst in Bytes."""
        memimg = bytearray(blen << size)
        if mem is not None:
            memimg = mem[(offs & ~mmask) : (offs & ~mmask) + (blen << size)]
        allstrb = (1 << (1 << size)) - 1
        if isinstance(wdata, int):
            wdata = iter((wdata,))
        else:
            wdata = iter(wdata)
        if wstrb is None:
            wstrb = iter(_bottomless(allstrb))
        elif isinstance(wstrb, int):
            wstrb = iter(_bottomless(wstrb))
        else:
            wstrb = iter(wstrb)
        for widx in range(blen):
            midx = (offs + (widx << size)) & mmask
            bytes = bytearray(int.to_bytes(next(wdata), 1 << size, "little"))
            strbs = next(wstrb)
            for b in range(1 << size):
                if strbs & 1:
                    memimg[midx + b] = bytes[b]
                strbs >>= 1
        return memimg

    def calc_expected(self, offs: int, size: SizeType, blen: int, mmask: int, mem: bytearray) -> list[int]:
        """Calculate Expected Read Data for a Burst according to Size."""
        xdata = []
        for widx in range(blen):
            xd = 0
            for bidx in range(1 << size):
                xd |= mem[(offs & ~mmask) + ((offs + (widx << size) + bidx) & mmask)] << (bidx << 3)
            xdata.append(xd)
        return xdata

    def _info_log(
        self,
        rw: str,
        addr: int,
        burst_type: BurstType,
        burst_length: int,
        size: SizeType,
        data: list[int],
        err_resp: bool,
    ) -> None:
        """Handle Master Logging."""
        if burst_type == BurstType.INCR:
            blenstr = f" burst length: {burst_length}"
        else:
            blenstr = ""
        err = " ERROR RESP" if err_resp else ""
        hexaddr = str(Hex(addr, self.addr_width))
        hexdata = ",".join(str(Hex(x, 8 << size)) for x in data)
        self.logger.info(
            f"=MST {rw}{err}= address: {hexaddr} data: [{hexdata}] "
            f"size: {size.name} burst: {burst_type.name}{blenstr}"
        )


class AHBSlaveDriver:
    """Active AHB Slave that can respond to Master requests."""

    def __init__(
        self,
        name: str,
        clk: SimHandle,
        rst_an: SimHandle,
        haddr: SimHandle,
        hwrite: SimHandle,
        hwdata: SimHandle,
        htrans: SimHandle,
        hburst: SimHandle,
        hsize: SimHandle,
        hrdata: SimHandle,
        hready: SimHandle,
        hreadyout: SimHandle,
        hresp: SimHandle,
        hsel: SimHandle,
        hprot: SimHandle = None,
        hwstrb: SimHandle = None,
        hnonsec: SimHandle = None,
        hmastlock: SimHandle = None,
        hexcl: SimHandle = None,
        hmaster: SimHandle = None,
        hexokay: SimHandle = None,
        hauser: SimHandle = None,
        hreadyout_delay: int = 0,
        size_bytes: int = 1024,
        err_addr: dict[Literal["r", "w", "rw"], list[int]] | None = None,
        log_level: int | None = None,
    ):
        """AHB Slave Init."""
        self.name = name
        self.clk = clk
        self.rst_an = rst_an
        self.haddr = haddr
        self.hwrite = hwrite
        self.hwdata = hwdata
        self.htrans = htrans
        self.hburst = hburst
        self.hsize = hsize
        self.hrdata = hrdata
        self.hready = hready
        self.hreadyout = hreadyout
        self.hresp = hresp
        self.hsel = hsel
        # optional signals:
        self.hprot = hprot
        self.hwstrb = hwstrb
        self.hnonsec = hnonsec
        self.hmastlock = hmastlock
        self.hexcl = hexcl
        self.hmaster = hmaster
        self.hexokay = hexokay
        self.hauser = hauser
        self.addr_width = len(haddr)
        self.data_width = len(hwdata)
        self.byte_width = len(hwdata) // 8
        self.err_addr = err_addr

        self.logger = getLogger(name)
        if log_level is not None:  # important explicit check for None as 0 would be a valid value
            self.logger.setLevel(log_level)

        self.mem = bytearray(size_bytes)  # Initialize a 1KB memory
        self.hreadyout_delay = hreadyout_delay  # Delay for HREADYOUT signal to simulate longer access times
        self.addrmask = size_bytes - 1

        # state variables
        self.state = 0
        self.burst_count = 0  # Burst count for burst transactions
        self.curr_addr = None
        self.curr_wdata = None
        self.curr_write = None
        self.curr_size = None
        self.curr_wstrb = None
        self.curr_hauser = None
        self.curr_hmaster = None

    def _read(self, addr: int, size: int) -> int:
        """AHB Read."""
        # number of bytes in this transfer according to transfer size
        byte_cnt = 2**size
        # extract the data from the bus
        alignmask = byte_cnt - 1
        shift_mask = self.data_width - 1
        datashift_bit = (addr << 3) & shift_mask
        masked_addr = self.addrmask & addr
        unaligned = alignmask & addr
        assert not unaligned, f"Address is unaligned for read with HSIZE of {SizeType(size)!r} at HADDR {addr}."

        rdata = int.from_bytes(self.mem[masked_addr : masked_addr + byte_cnt], "little") << datashift_bit
        hexaddr = str(Hex(addr, self.addr_width))
        hexdata = str(Hex(rdata, 8 << size))
        self.logger.info(f"=SLV READ= address: {hexaddr} data: {hexdata} size: {SizeType(size).name}")
        self.logger.debug(
            f"=SLV READ= alignment mask: {hex(alignmask)} "
            f"shift mask: {hex(shift_mask)} datashift in bits: {hex(datashift_bit)} "
            f"address (masked): {hex(masked_addr)} byte count: {byte_cnt}"
        )
        return rdata

    def _write(self, addr: int, size: int, data: int, strb: int) -> None:
        """AHB Write."""
        # number of bytes in this transfer according to transfer size
        byte_cnt = 2**size
        shift_mask = self.data_width - 1
        # extract the data from the bus
        alignmask = byte_cnt - 1
        lower_datamask = (2 ** (byte_cnt * 8)) - 1
        datashift_bit = (addr << 3) & shift_mask
        unaligned = alignmask & addr
        assert not unaligned, f"Address is unaligned for write with HSIZE of {size} at HADDR {addr}."

        masked_addr = self.addrmask & addr
        wdata = (data >> datashift_bit) & lower_datamask
        wstrb = strb >> (datashift_bit >> 3)
        strbs = [(wstrb >> i) & 1 for i in range(byte_cnt)]
        bytes = bytearray(int.to_bytes(wdata, byte_cnt, "little"))
        for b in range(byte_cnt):
            if not strbs[b]:
                bytes[b] = self.mem[masked_addr + b]  # replace with read value

        hexaddr = str(Hex(addr, self.addr_width))
        hexdata = str(Hex(wdata, 8 << size))
        self.logger.info(f"=SLV WRITE= address: {hexaddr} data: {hexdata} size: {SizeType(size).name} strb: {strbs}")
        self.logger.debug(
            f"=SLV WRITE= alignment mask: {hex(alignmask)} shift mask: {hex(shift_mask)} "
            f"lower data mask {hex(lower_datamask)} datashift in bits: {hex(datashift_bit)} "
            f"address (masked): {hex(masked_addr)} byte count: {byte_cnt} "
            f"data (bytes): {','.join([hex(x) for x in bytes])} "
        )
        self.mem[masked_addr : masked_addr + byte_cnt] = bytes

    def _check_err_addr(self, haddr: int, hwrite: int) -> bool:
        """Check for Error Address."""
        if self.err_addr is None:
            return False
        if hwrite:
            if (haddr in self.err_addr.get("w", [])) or (haddr in self.err_addr.get("rw", [])):
                return True
        elif (haddr in self.err_addr.get("r", [])) or (haddr in self.err_addr.get("rw", [])):
            return True
        return False

    async def run(self):
        """Slave Main Loop."""
        self.hreadyout.value = 1
        self.hrdata.value = 0xDEADDEAD
        self.hresp.value = 0
        allstrb = (1 << self.byte_width) - 1
        while True:
            await RisingEdge(self.clk)
            if self.state:
                for _ in range(self.hreadyout_delay):  # delay the answer if configured
                    await RisingEdge(self.clk)
                    self.hreadyout.value = 0
                self.hreadyout.value = 1
                self.curr_wdata = self.hwdata.value if self.curr_write else 0
                self.curr_wstrb = self.hwstrb.value.integer if self.hwstrb is not None else allstrb
                if (self.state == 1) and self.curr_write:
                    # Handle write request
                    self._write(self.curr_addr, self.curr_size, self.curr_wdata, self.curr_wstrb)
            # Check if there's an AHB request
            if self.hsel.value and self.htrans.value in (TransType.SEQ, TransType.NONSEQ):
                self.curr_addr = self.haddr.value.integer
                self.curr_trans = self.htrans.value.integer
                self.curr_write = self.hwrite.value.integer
                self.curr_size = self.hsize.value.integer
                self.curr_hauser = self.hauser.value.integer if self.hauser is not None else None
                self.curr_hmaster = self.hmaster.value.integer if self.hmaster is not None else None
                self.state = 1
            else:
                self.state = 0
            if self.state and self._check_err_addr(self.curr_addr, self.curr_write):
                acc = "WRITE" if self.curr_write else "READ"
                hexaddr = str(Hex(self.curr_addr, self.addr_width))
                self.logger.info(f"=SLV ERROR RESP for {acc}= address: {hexaddr}")
                # TODO: response handling w/ delay cycle...
                self.state = 2
            if self.state == 1 and self.curr_trans == TransType.NONSEQ:
                acc = "WRITE" if self.curr_write else "READ"
                hauser = f" hauser: {hex(self.curr_hauser)}" if self.curr_hauser is not None else ""
                hmaster = f" hmaster: {hex(self.curr_hmaster)}" if self.curr_hmaster is not None else ""
                hexaddr = str(Hex(self.curr_addr, self.addr_width))
                self.logger.info(f"=SLV NEW {acc}= address: {hexaddr}{hauser}{hmaster}")
            if self.state and not self.curr_write:
                # Handle read request (need to apply read value for next cycle)
                rdata = self._read(self.curr_addr, self.curr_size)
                self.hrdata.value = rdata
            else:
                self.hrdata.value = 0xDEADDEAD

    def set_hreadyout_delay(self, delay):
        """Set hreadyout Delay."""
        self.hreadyout_delay = delay

    def set_data(self, data):
        """Preload Slave Memory."""
        self.mem = data

    def log_data(self, start_addr=0, end_addr=None, chunk_size=16):
        """Request logging of Slave Memory content."""
        if end_addr is None:
            end_addr = len(self.mem) - 1
        elif end_addr > len(self.mem) - 1:
            self.logger.error(
                f"=MEMORY CONTENTS= Provided end_addr {hex(end_addr)} to log_data is beyond size of slave memory."
            )
            return
        for i in range(start_addr, end_addr + 1, chunk_size):
            hexstart = str(Hex(i, self.addr_width))
            hexend = str(Hex(i + chunk_size - 1, self.addr_width))
            chunk = ",".join(str(Hex(x, 8)) for x in self.mem[i : min(i + chunk_size, end_addr + 1)])
            self.logger.info(f"=MEMORY CONTENTS= {hexstart}-{hexend} [{chunk}]")



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, mowithcommentestriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and toublicommentistribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""

AMBA Type Definitions.

Mainly these types are needed:

* :any:`AhbMstType`
* :any:`AhbSlvType`
* :any:`ApbSlvType`
"""

from logging import getLogger

import ucdp as u

LOGGER = getLogger(__name__)

LEGAL_AHB_ADDR_WIDTH = range(10, 65)
LEGAL_AHB_DATA_WIDTH = [8, 16, 32, 64, 128, 256, 512, 1024]
LEGAL_AHB_PROT_WIDTH = [0, 4, 7]
LEGAL_APB_ADDR_WIDTH = range(33)
LEGAL_APB_DATA_WIDTH = [8, 16, 32]


class ASecIdType(u.AEnumType):
    """
    Base of all Security ID Types.
    """

    keytype: u.UintType = u.UintType(4)

    def get_subset_type(self, excludes: u.Names | None = None):
        """Return variant with element excluded."""
        excludes = u.split(excludes)
        return self.new(filter_=lambda item: item.value not in excludes)


class AhbProtType(u.UintType):
    """
    AHB Protection.

    >>> AhbProtType().width
    4
    >>> AhbProtType(width=7).width
    7

    Width is checked for legal values:

    >>> t = AhbProtType(width=5)
    Traceback (most recent call last):
    ...
    ValueError: Illegal value for AHB hprotwidth: 5. Legal values are [0, 4, 7]
    """

    title: str = "AHB Transfer Protection"
    comment: str = "AHB Transfer Protection"

    def __init__(self, width=4, **kwargs):
        if width not in LEGAL_AHB_PROT_WIDTH:
            raise ValueError(f"Illegal value for AHB hprotwidth: {width}. Legal values are {LEGAL_AHB_PROT_WIDTH}")
        super().__init__(width=width, default=3, **kwargs)


class AhbHMastType(u.UintType):
    """
    AHB Master ID for Exclusive Transfers.
    """

    title: str = "AHB Master ID"
    comment: str = "AHB Master ID"

    def __init__(self, width=4, **kwargs):
        super().__init__(width=width, **kwargs)


class AmbaProto(u.AConfig):
    """
    Amba Protocol Version.

    Specification of optional/conditional Protocol Features.
    Used for both, AHB and APB. However, features not defined for APB are simply ignored.

    hprotwidth (int): Bitwidth for AHB hprot signal
    has_hburst(bool): Determines whether AHB hburst signal exists (otherwise INCR only)
    has_hmastlock (bool): Determines whether AHB hmastlock exists
    has_hnonsec (bool): Determines whether AHB hnonsec signal exists
    has_exclxfers (bool): Determines whether AHB signals for exclusive transfers exist (hexcl, hmaster, hexokay)
    hmaster_width (int): Bitwidth for AHB hmaster signal
    enh_hmaster )bool): Determines whether hmaster signal is enhanced with master index on an interconnect
    has_wstrb (bool): Determines whether AHB hwstrb or APB pstrb write strobe signals exist
    has_pprot (bool): Determines whether APB pprot signal exists
    has_pnse (bool):  Determines whether APB uses Realm Management Extensions
    ausertype (AEnumType|UintType): Type used for AHB hauser or APB pauser
    wusertype (AEnumType|UintType): Type used for AHB hwuser or APB pwuser signal
    rusertype (AEnumType|UintType): Type used for AHB hruser or APB pruser signal
    busertype (AEnumType|UintType): Type used for AHB hbuser or APB pbuser signal
    """

    hprotwidth: int = 4
    hmaster_width: int = 0
    enh_hmaster: bool = False
    has_hburst: bool = True
    has_hmastlock: bool = False
    has_hnonsec: bool = False
    has_exclxfers: bool = False
    has_wstrb: bool = False
    has_pprot: bool = False
    has_pnse: bool = False
    ausertype: u.AEnumType | u.UintType | None = None
    wusertype: u.AEnumType | u.UintType | None = None
    rusertype: u.AEnumType | u.UintType | None = None
    busertype: u.AEnumType | u.UintType | None = None

    @property
    def hprottype(self) -> AhbProtType | None:
        """Protocol has HPROT signal."""
        if self.hprotwidth:
            return AhbProtType(width=self.hprotwidth)
        return None

    @property
    def hmaster_type(self) -> AhbHMastType | None:
        """Protocol has HMASTER signal."""
        if self.hmaster_width:
            return AhbHMastType(width=self.hmaster_width)
        return None


AMBA3 = AmbaProto(name="amba3")


class AhbMstType(u.AStructType):
    """
    From AHB Master.

    Keyword Args:
        proto (Protocol): Protocol feature set selection.

    The default type:

    >>> for item in AhbMstType().values(): item
    StructItem('htrans', AhbTransType(), doc=Doc(title='AHB Transfer Type', ...))
    StructItem('haddr', AhbAddrType(32), doc=Doc(title='AHB Bus Address', ...))
    StructItem('hwrite', AhbWriteType(), doc=Doc(title='AHB Write Enable', ...))
    StructItem('hsize', AhbSizeType(), doc=Doc(title='AHB Size', ...))
    StructItem('hburst', AhbBurstType(), doc=Doc(title='AHB Burst Type', ...))
    StructItem('hprot', AhbProtType(4, default=3), doc=Doc(title='AHB Transfer Protection', ...))
    StructItem('hwdata', AhbDataType(32), doc=Doc(title='AHB Data', ...))
    StructItem('hready', AhbReadyType(), orientation=BWD, doc=Doc(title='AHB Transfer Done', ...))
    StructItem('hresp', AhbRespType(), orientation=BWD, doc=Doc(title='AHB Response Error', ...))
    StructItem('hrdata', AhbDataType(32), orientation=BWD, doc=Doc(title='AHB Data', ...))

    With all extras:

    >>> class AuserType(ASecIdType):
    ...     def _build(self):
    ...         self._add(0, "apps")
    ...         self._add(2, "comm")
    ...         self._add(5, "audio")
    >>> ahb5 = AmbaProto("AHB5", ausertype=AuserType(default=2), hmaster_width=3,\
    has_hmastlock=True, has_hnonsec=True, has_exclxfers=True, has_wstrb=True)

    >>> for item in AhbMstType(proto=ahb5).values(): item
    StructItem('htrans', AhbTransType(), doc=Doc(title='AHB Transfer Type', ...))
    StructItem('haddr', AhbAddrType(32), doc=Doc(title='AHB Bus Address', ...))
    StructItem('hauser', AuserType(default=2), doc=Doc(title='AHB Address User Channel', ...))
    StructItem('hwrite', AhbWriteType(), doc=Doc(title='AHB Write Enable', ...))
    StructItem('hsize', AhbSizeType(), doc=Doc(title='AHB Size', ...))
    StructItem('hburst', AhbBurstType(), doc=Doc(title='AHB Burst Type', ...))
    StructItem('hprot', AhbProtType(4, default=3), doc=Doc(title='AHB Transfer Protection', ...))
    StructItem('hnonsec', AhbNonsecType(), doc=Doc(title='AHB Secure Transfer', ...))
    StructItem('hmastlock', AhbMastlockType(), doc=Doc(title='AHB Locked Sequence Enable', ...))
    StructItem('hwdata', AhbDataType(32), doc=Doc(title='AHB Data', ...))
    StructItem('hwstrb', AhbWstrbType(4), doc=Doc(title='AHB Write Strobe', ...))
    StructItem('hexcl', AhbExclType(), doc=Doc(title='AHB Exclusive Transfer', ...))
    StructItem('hmaster', AhbHMastType(3), doc=Doc(title='AHB Master ID', ...))
    StructItem('hready', AhbReadyType(), orientation=BWD, doc=Doc(title='AHB Transfer Done', ...))
    StructItem('hresp', AhbRespType(), orientation=BWD, doc=Doc(title='AHB Response Error', ...))
    StructItem('hexokay', AhbHexokType(), orientation=BWD, doc=Doc(title='AHB Exclusive Response', ...))
    StructItem('hrdata', AhbDataType(32), orientation=BWD, doc=Doc(title='AHB Data', ...))

    Both protocol versions are not connectable:

    >>> AhbMstType().is_connectable(AhbMstType(proto=ahb5))
    False

    But casting is allowed:

    >>> for item in AhbMstType().cast(AhbMstType(proto=ahb5)): item
    ('', '')
    ('htrans', 'htrans')
    ('haddr', 'haddr')
    ('hwrite', 'hwrite')
    ('hsize', 'hsize')
    ('hburst', 'hburst')
    ('hprot', 'hprot')
    ('hwdata', 'hwdata')
    ('hready', 'hready')
    ('hresp', 'hresp')
    ('hrdata', 'hrdata')

    It is also not allowed to connect Master and Slave:

    >>> AhbMstType().is_connectable(AhbSlvType())
    False

    But casting is allowed in all ways.

    >>> len(tuple(AhbMstType().cast(AhbSlvType())))
    12
    >>> len(tuple(AhbMstType(proto=ahb5).cast(AhbSlvType())))
    11
    >>> len(tuple(AhbMstType(proto=ahb5).cast(AhbSlvType(proto=ahb5))))
    12

    Without HPROT:

    >>> ahbp = AmbaProto("AHB3", hprotwidth=0)

    >>> for item in AhbMstType(proto=ahbp).values(): item
    StructItem('htrans', AhbTransType(), doc=Doc(title='AHB Transfer Type', ...))
    StructItem('haddr', AhbAddrType(32), doc=Doc(title='AHB Bus Address', ...))
    StructItem('hwrite', AhbWriteType(), doc=Doc(title='AHB Write Enable', ...))
    StructItem('hsize', AhbSizeType(), doc=Doc(title='AHB Size', ...))
    StructItem('hburst', AhbBurstType(), doc=Doc(title='AHB Burst Type', ...))
    StructItem('hwdata', AhbDataType(32), doc=Doc(title='AHB Data', ...))
    StructItem('hready', AhbReadyType(), orientation=BWD, doc=Doc(title='AHB Transfer Done', ...))
    StructItem('hresp', AhbRespType(), orientation=BWD, doc=Doc(title='AHB Response Error', ...))
    StructItem('hrdata', AhbDataType(32), orientation=BWD, doc=Doc(title='AHB Data', ...))
    """

    title: str = "AHB Master"
    comment: str = "AHB Master"
    proto: AmbaProto = AMBA3
    addrwidth: int = 32
    datawidth: int = 32

    def _build(self):  # noqa: C901
        # FWD
        self._add("htrans", AhbTransType())
        self._add("haddr", AhbAddrType(self.addrwidth))
        if ausertype := self.proto.ausertype:
            auserdoc = "AHB Address User Channel"
            self._add("hauser", ausertype, title=auserdoc, comment=auserdoc)
        if wusertype := self.proto.wusertype:
            wuserdoc = "AHB Write Data User Channel"
            self._add("hwuser", wusertype, title=wuserdoc, comment=wuserdoc)
        self._add("hwrite", AhbWriteType())
        self._add("hsize", AhbSizeType())
        if self.proto.has_hburst:
            self._add("hburst", AhbBurstType())
        if hprottype := self.proto.hprottype:
            self._add("hprot", hprottype)
        if self.proto.has_hnonsec:
            self._add("hnonsec", AhbNonsecType())
        if self.proto.has_hmastlock:
            self._add("hmastlock", AhbMastlockType())
        self._add("hwdata", AhbDataType(self.datawidth))
        if self.proto.has_wstrb:
            self._add("hwstrb", AhbWstrbType(self.datawidth))
        if self.proto.has_exclxfers:
            self._add("hexcl", AhbExclType())
        if hmasttype := self.proto.hmaster_type:
            self._add("hmaster", hmasttype)
        # BWD
        self._add("hready", AhbReadyType(), u.BWD)
        self._add("hresp", AhbRespType(), u.BWD)
        if self.proto.has_exclxfers:
            self._add("hexokay", AhbHexokType(), u.BWD)
        self._add("hrdata", AhbDataType(self.datawidth), u.BWD)
        if rusertype := self.proto.rusertype:
            ruserdoc = "AHB Read Data User Channel"
            self._add("hruser", rusertype, title=ruserdoc, comment=ruserdoc, orientation=u.BWD)
        if busertype := self.proto.busertype:
            buserdoc = "AHB Read Response User Channel"
            self._add("hbuser", busertype, title=buserdoc, comment=buserdoc, orientation=u.BWD)

    def cast(self, other):
        """
        How to cast an assign of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, AhbMstType) and self.proto != other.proto:
            # Drive a Mst with Mst signals
            yield "", ""
            yield "htrans", "htrans"
            yield "haddr", "haddr"
            # never use hauser
            yield "hwrite", "hwrite"
            yield "hsize", "hsize"
            yield "hburst", "hburst"
            yield "hprot", "hprot"
            yield "hwdata", "hwdata"
            # BWD
            yield "hready", "hready"
            yield "hresp", "hresp"
            yield "hrdata", "hrdata"
        elif isinstance(other, AhbSlvType):
            # Drive a Mst with Slv signals
            yield "", ""
            yield "htrans", "htrans"
            yield "haddr", "haddr"
            if self.proto.ausertype == other.proto.ausertype:
                yield "hauser", "hauser"
            yield "hwrite", "hwrite"
            yield "hsize", "hsize"
            yield "hburst", "hburst"
            yield "hprot", "hprot"
            yield "hwdata", "hwdata"
            # BWD
            yield "hready", "hreadyout"
            # yield "hready", "hready" TODO: @djakschik, extend cast support
            yield "hresp", "hresp"
            yield "hrdata", "hrdata"
        else:
            return None


class AhbSlvType(u.AStructType):
    """
    To AHB Slave.

    Keyword Args:
        proto (Protocol): Protocol feature set selection.

    The default type:

    >>> for item in AhbSlvType().values(): item
    StructItem('hsel', AhbSelType(), doc=Doc(title='AHB Slave Select', ...))
    StructItem('haddr', AhbAddrType(32), doc=Doc(title='AHB Bus Address', ...))
    StructItem('hwrite', AhbWriteType(), doc=Doc(title='AHB Write Enable', ...))
    StructItem('htrans', AhbTransType(), doc=Doc(title='AHB Transfer Type', ...))
    StructItem('hsize', AhbSizeType(), doc=Doc(title='AHB Size', ...))
    StructItem('hburst', AhbBurstType(), doc=Doc(title='AHB Burst Type', ...))
    StructItem('hprot', AhbProtType(4, default=3), doc=Doc(title='AHB Transfer Protection', ...))
    StructItem('hwdata', AhbDataType(32), doc=Doc(title='AHB Data', ...))
    StructItem('hready', AhbReadyType(), doc=Doc(title='AHB Transfer Done to Slave', ...))
    StructItem('hreadyout', AhbReadyType(), orientation=BWD, doc=Doc(title='AHB Transfer Done from Slave', ...))
    StructItem('hresp', AhbRespType(), orientation=BWD, doc=Doc(title='AHB Response Error', ...))
    StructItem('hrdata', AhbDataType(32), orientation=BWD, doc=Doc(title='AHB Data', ...))

    With `hauser`:

    >>> class AuserType(ASecIdType):
    ...     def _build(self):
    ...         self._add(0, "apps")
    ...         self._add(2, "comm")
    ...         self._add(5, "audio")
    >>> ahb5 = AmbaProto("AHB5", ausertype=AuserType(default=2))

    >>> for item in AhbSlvType(proto=ahb5).values(): item
    StructItem('hsel', AhbSelType(), doc=Doc(title='AHB Slave Select', ...))
    StructItem('haddr', AhbAddrType(32), doc=Doc(title='AHB Bus Address', ...))
    StructItem('hauser', AuserType(default=2), doc=Doc(title='AHB Address User Channel', ...))
    StructItem('hwrite', AhbWriteType(), doc=Doc(title='AHB Write Enable', ...))
    StructItem('htrans', AhbTransType(), doc=Doc(title='AHB Transfer Type', ...))
    StructItem('hsize', AhbSizeType(), doc=Doc(title='AHB Size', ...))
    StructItem('hburst', AhbBurstType(), doc=Doc(title='AHB Burst Type', ...))
    StructItem('hprot', AhbProtType(4, default=3), doc=Doc(title='AHB Transfer Protection', ...))
    StructItem('hwdata', AhbDataType(32), doc=Doc(title='AHB Data', ...))
    StructItem('hready', AhbReadyType(), doc=Doc(title='AHB Transfer Done to Slave', ...))
    StructItem('hreadyout', AhbReadyType(), orientation=BWD, doc=Doc(title='AHB Transfer Done from Slave', ...))
    StructItem('hresp', AhbRespType(), orientation=BWD, doc=Doc(title='AHB Response Error', ...))
    StructItem('hrdata', AhbDataType(32), orientation=BWD, doc=Doc(title='AHB Data', ...))

    Both protocol versions are not connectable:

    >>> AhbSlvType().is_connectable(AhbSlvType(proto=ahb5))
    False

    But casting is allowed:

    >>> for item in AhbSlvType().cast(AhbSlvType(proto=ahb5)): item
    ('', '')
    ('hsel', 'hsel')
    ('haddr', 'haddr')
    ('hwrite', 'hwrite')
    ('htrans', 'htrans')
    ('hsize', 'hsize')
    ('hburst', 'hburst')
    ('hprot', 'hprot')
    ('hwdata', 'hwdata')
    ('hready', 'hready')
    ('hreadyout', 'hreadyout')
    ('hresp', 'hresp')
    ('hrdata', 'hrdata')

    It is also not allowed to connect Master and Slave:

    >>> AhbSlvType().is_connectable(AhbMstType())
    False

    But casting is allowed in all ways.

    >>> len(tuple(AhbSlvType().cast(AhbMstType())))
    14
    >>> len(tuple(AhbSlvType(proto=ahb5).cast(AhbMstType())))
    13
    >>> len(tuple(AhbSlvType(proto=ahb5).cast(AhbMstType(proto=ahb5))))
    14

    With wider HPROT:

    >>> ahbp = AmbaProto("AHB5", hprotwidth=7)

    >>> for item in AhbSlvType(proto=ahbp).values(): item
    StructItem('hsel', AhbSelType(), doc=Doc(title='AHB Slave Select', ...))
    StructItem('haddr', AhbAddrType(32), doc=Doc(title='AHB Bus Address', ...))
    StructItem('hwrite', AhbWriteType(), doc=Doc(title='AHB Write Enable', ...))
    StructItem('htrans', AhbTransType(), doc=Doc(title='AHB Transfer Type', ...))
    StructItem('hsize', AhbSizeType(), doc=Doc(title='AHB Size', ...))
    StructItem('hburst', AhbBurstType(), doc=Doc(title='AHB Burst Type', ...))
    StructItem('hprot', AhbProtType(7, default=3), doc=Doc(title='AHB Transfer Protection', ...))
    StructItem('hwdata', AhbDataType(32), doc=Doc(title='AHB Data', ...))
    StructItem('hready', AhbReadyType(), doc=Doc(title='AHB Transfer Done to Slave', ...))
    StructItem('hreadyout', AhbReadyType(), orientation=BWD, doc=Doc(title='AHB Transfer Done from Slave', ...))
    StructItem('hresp', AhbRespType(), orientation=BWD, doc=Doc(title='AHB Response Error', ...))
    StructItem('hrdata', AhbDataType(32), orientation=BWD, doc=Doc(title='AHB Data', ...))

    """

    title: str = "AHB Slave"
    comment: str = "AHB Slave"
    proto: AmbaProto = AMBA3
    addrwidth: int = 32
    datawidth: int = 32

    def _build(self):  # noqa: C901
        # FWD
        self._add("hsel", AhbSelType())
        self._add("haddr", AhbAddrType(self.addrwidth))
        if ausertype := self.proto.ausertype:
            auserdoc = "AHB Address User Channel"
            self._add("hauser", ausertype, title=auserdoc, comment=auserdoc)
        if wusertype := self.proto.wusertype:
            wuserdoc = "AHB Write Data User Channel"
            self._add("hwuser", wusertype, title=wuserdoc, comment=wuserdoc)
        self._add("hwrite", AhbWriteType())
        self._add("htrans", AhbTransType())
        self._add("hsize", AhbSizeType())
        if self.proto.has_hburst:
            self._add("hburst", AhbBurstType())
        if self.proto.hprottype:
            self._add("hprot", self.proto.hprottype)
        if self.proto.has_hnonsec:
            self._add("hnonsec", AhbNonsecType())
        if self.proto.has_hmastlock:
            self._add("hmastlock", AhbMastlockType())
        self._add("hwdata", AhbDataType(self.datawidth))
        if self.proto.has_wstrb:
            self._add("hwstrb", AhbWstrbType(self.datawidth))
        title: str = "AHB Transfer Done to Slave"
        self._add("hready", AhbReadyType(), title=title, comment=title)
        if self.proto.has_exclxfers:
            self._add("hexcl", AhbExclType())
        if hmasttype := self.proto.hmaster_type:
            self._add("hmaster", hmasttype)
        # BWD
        title: str = "AHB Transfer Done from Slave"
        self._add("hreadyout", AhbReadyType(), u.BWD, title=title, comment=title)
        self._add("hresp", AhbRespType(), u.BWD)
        if self.proto.has_exclxfers:
            self._add("hexokay", AhbHexokType(), u.BWD)
        self._add("hrdata", AhbDataType(self.datawidth), u.BWD)
        if rusertype := self.proto.rusertype:
            ruserdoc = "AHB Read Data User Channel"
            self._add("hruser", rusertype, title=ruserdoc, comment=ruserdoc, orientation=u.BWD)
        if busertype := self.proto.busertype:
            buserdoc = "AHB Read Response User Channel"
            self._add("hbuser", busertype, title=buserdoc, comment=buserdoc, orientation=u.BWD)

    def cast(self, other):
        """
        How to cast an assign of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, AhbSlvType) and self.proto != other.proto:
            # Drive a Slv with Slv signals
            yield "", ""
            yield "hsel", "hsel"
            yield "haddr", "haddr"
            # never use hauser
            yield "hwrite", "hwrite"
            yield "htrans", "htrans"
            yield "hsize", "hsize"
            yield "hburst", "hburst"
            yield "hprot", "hprot"
            yield "hwdata", "hwdata"
            yield "hready", "hready"
            # BWD
            yield "hreadyout", "hreadyout"
            yield "hresp", "hresp"
            yield "hrdata", "hrdata"
        elif isinstance(other, AhbMstType):
            # Drive a Slv with Mst signals
            yield "", ""
            yield "hsel", "ternary(htrans > '1b0', '1b1', '1b0')"
            yield "haddr", "haddr"
            if self.proto.ausertype == other.proto.ausertype:
                yield "hauser", "hauser"
            yield "hwrite", "hwrite"
            yield "htrans", "htrans"
            yield "hsize", "hsize"
            yield "hburst", "hburst"
            yield "hprot", "hprot"
            yield "hwdata", "hwdata"
            yield "hready", u.const("1'b1")
            # BWD
            yield "hreadyout", "hready"
            yield "hresp", "hresp"
            yield "hrdata", "hrdata"
        else:
            return None


class AhbSelType(u.BitType):
    """AHB Select."""

    title: str = "AHB Slave Select"
    comment: str = "AHB Slave Select"


class AhbAddrType(u.UintType):
    """
    Address.

    >>> AhbAddrType().width
    32
    >>> AhbAddrType(width=16).width
    16

    Checked for valid width:
    >>> t = AhbAddrType(width=9)
    Traceback (most recent call last):
    ...
    ValueError: Illegal value for AHB addrwidth: 9.Legal values are between 10 and 64.
    """

    title: str = "AHB Bus Address"
    comment: str = "AHB Bus Address"

    def __init__(self, width=32, **kwargs):
        if width not in LEGAL_AHB_ADDR_WIDTH:
            raise ValueError(
                f"Illegal value for AHB addrwidth: {width}."
                f"Legal values are between {LEGAL_AHB_ADDR_WIDTH[0]} and {LEGAL_AHB_ADDR_WIDTH[-1]}."
            )
        super().__init__(width=width, **kwargs)


class AhbDataType(u.UintType):
    """
    AHB Data.

    >>> AhbDataType().width
    32
    >>> AhbDataType(width=64).width
    64

    Data width is checked for legal values:

    >>> t = AhbDataType(width=57)
    Traceback (most recent call last):
    ...
    ValueError: Illegal value for AHB datawidth: 57. Legal values are [8, 16, 32, 64, 128, 256, 512, 1024]
    """

    title: str = "AHB Data"
    comment: str = "AHB Data"

    def __init__(self, width=32, **kwargs):
        if width not in LEGAL_AHB_DATA_WIDTH:
            raise ValueError(f"Illegal value for AHB datawidth: {width}. Legal values are {LEGAL_AHB_DATA_WIDTH}")
        super().__init__(width=width, **kwargs)


class AhbTransType(u.AEnumType):
    """
    AHB Transfer Type.

    >>> for item in AhbTransType().values(): item
    EnumItem(0, 'idle', doc=Doc(title='No transfer'))
    EnumItem(1, 'busy', doc=Doc(title='Idle cycle within transfer'))
    EnumItem(2, 'nonseq', doc=Doc(title='Single transfer or first transfer of a burst'))
    EnumItem(3, 'seq', doc=Doc(title='Consecutive transfers of a burst'))
    """

    keytype: u.UintType = u.UintType(2)
    title: str = "AHB Transfer Type"
    comment: str = "AHB Transfer Type"

    def _build(self):
        self._add(0, "idle", "No transfer")
        self._add(1, "busy", "Idle cycle within transfer")
        self._add(2, "nonseq", "Single transfer or first transfer of a burst")
        self._add(3, "seq", "Consecutive transfers of a burst")


class AhbSizeType(u.AEnumType):
    """
    AHB Size Type.

    >>> for item in AhbSizeType().values(): item
    EnumItem(0, 'byte', doc=Doc(title='Byte', descr='8 bits'))
    EnumItem(1, 'halfword', doc=Doc(title='Halfword', descr='16 bits'))
    EnumItem(2, 'word', doc=Doc(title='Word', descr='32 bits'))
    EnumItem(3, 'doubleword', doc=Doc(title='Doubleword', descr='64 bits'))
    ...
    """

    keytype: u.UintType = u.UintType(3)
    title: str = "AHB Size"
    comment: str = "AHB Size"

    def _build(self):
        self._add(0, "byte", "Byte", descr="8 bits")
        self._add(1, "halfword", "Halfword", descr="16 bits")
        self._add(2, "word", "Word", descr="32 bits")
        self._add(3, "doubleword", "Doubleword", descr="64 bits")
        self._add(4, "fourword", "4-word", descr="128 bits")
        self._add(5, "eightword", "8-word", descr="256 bits")
        self._add(6, "sixteenword", "16-word", descr="512 bits")
        self._add(7, "kilobit", "32-word", descr="1024 bits")


class AhbBurstType(u.AEnumType):
    """
    AHB Burst Type.

    >>> for item in AhbBurstType().values(): item
    EnumItem(0, 'single', doc=Doc(title='Single transfer'))
    EnumItem(1, 'incr', doc=Doc(title='Incrementing burst of unspecified length'))
    EnumItem(2, 'wrap4', doc=Doc(title='4-beat wrapping burst'))
    EnumItem(3, 'incr4', doc=Doc(title='4-beat incrementing burst'))
    EnumItem(4, 'wrap8', doc=Doc(title='8-beat wrapping burst'))
    EnumItem(5, 'incr8', doc=Doc(title='8-beat incrementing burst'))
    EnumItem(6, 'wrap16', doc=Doc(title='16-beat wrapping burst'))
    EnumItem(7, 'incr16', doc=Doc(title='16-beat incrementing burst'))
    """

    keytype: u.UintType = u.UintType(3)
    title: str = "AHB Burst Type"
    comment: str = "AHB Burst Type"

    def _build(self):
        self._add(0, "single", "Single transfer")
        self._add(1, "incr", "Incrementing burst of unspecified length")
        self._add(2, "wrap4", "4-beat wrapping burst")
        self._add(3, "incr4", "4-beat incrementing burst")
        self._add(4, "wrap8", "8-beat wrapping burst")
        self._add(5, "incr8", "8-beat incrementing burst")
        self._add(6, "wrap16", "16-beat wrapping burst")
        self._add(7, "incr16", "16-beat incrementing burst")


class AhbWriteType(u.AEnumType):
    """
    AHB Write Type.

    >>> for item in AhbWriteType().values(): item
    EnumItem(0, 'read', doc=Doc(title='Read operation'))
    EnumItem(1, 'write', doc=Doc(title='Write operation'))
    """

    keytype: u.BitType = u.BitType()
    title: str = "AHB Write Enable"
    comment: str = "AHB Write Enable"

    def _build(self):
        self._add(0, "read", "Read operation")
        self._add(1, "write", "Write operation")


class AhbWstrbType(u.UintType):
    """
    AHB Write Strobe.

    Write strobe width is derived directly from data width.
    >>> AhbWstrbType().width
    4

    Data width is checked for legal values:
    >>> t = AhbWstrbType(42)
    Traceback (most recent call last):
    ...
    ValueError: Illegal value for AHB datawidth: 42. Legal values are [8, 16, 32, 64, 128, 256, 512, 1024]
    """

    title: str = "AHB Write Strobe"
    comment: str = "AHB Write Strobe"

    def __init__(self, datawidth=32, **kwargs):
        if datawidth not in LEGAL_AHB_DATA_WIDTH:
            raise ValueError(f"Illegal value for AHB datawidth: {datawidth}. Legal values are {LEGAL_AHB_DATA_WIDTH}")
        super().__init__(width=datawidth // 8, **kwargs)


class AhbNonsecType(u.AEnumType):
    """
    AHB Secure Transfer Type.
    """

    keytype: u.BitType = u.BitType()
    title: str = "AHB Secure Transfer"
    comment: str = "AHB Secure Transfer"

    def _build(self):
        self._add(0, "sec", "Secure Transfer")
        self._add(1, "nonsec", "Nonsecure Transfer")


class AhbMastlockType(u.AEnumType):
    """
    AHB Master Lock Type.
    """

    keytype: u.BitType = u.BitType()
    title: str = "AHB Locked Sequence Enable"
    comment: str = "AHB Locked Sequence Enable"

    def _build(self):
        self._add(0, "norm", "Normal sequence")
        self._add(1, "locked", "Locked sequence")


class AhbRespType(u.AEnumType):
    """
    AHB Response Type.

    >>> for item in AhbRespType().values(): item
    EnumItem(0, 'okay', doc=Doc(title='OK'))
    EnumItem(1, 'error', doc=Doc(title='Error'))
    """

    keytype: u.BitType = u.BitType()
    title: str = "AHB Response Error"
    comment: str = "AHB Response Error"

    def _build(self):
        self._add(0, "okay", "OK")
        self._add(1, "error", "Error")


class AhbHexokType(u.AEnumType):
    """AHB Hexokay Type."""

    keytype: u.BitType = u.BitType()
    title: str = "AHB Exclusive Response"
    comment: str = "AHB Exclusive Response"

    def _build(self):
        self._add(0, "error", "Error")
        self._add(1, "okay", "OK")


class AhbReadyType(u.AEnumType):
    """
    AHB Ready Type.

    >>> for item in AhbReadyType().values(): item
    EnumItem(0, 'busy', doc=Doc(title='Ongoing'))
    EnumItem(1, 'done', doc=Doc(title='Done'))
    """

    keytype: u.BitType = u.BitType(default=1)
    title: str = "AHB Transfer Done"
    comment: str = "AHB Transfer Done"

    def _build(self):
        self._add(0, "busy", "Ongoing")
        self._add(1, "done", "Done")


class AhbExclType(u.AEnumType):
    """
    AHB Exclusive Transfer Type.
    """

    keytype: u.BitType = u.BitType(default=1)
    title: str = "AHB Exclusive Transfer"
    comment: str = "AHB Exclusive Transfer"

    def _build(self):
        self._add(0, "norm", "Normal")
        self._add(1, "excl", "Exclusive")


################################################################################################
##   APB PRotocol Related
################################################################################################


class ApbSlvType(u.AStructType):
    """
    To APB Slave.

    Keyword Args:
        proto (Protocol): Protocol feature set selection.

    The default type:

    >>> for item in ApbSlvType().values(): item
    StructItem('paddr', ApbAddrType(12), doc=Doc(title='APB Bus Address', ...))
    StructItem('pwrite', ApbWriteType(), doc=Doc(title='APB Write Enable', ...))
    StructItem('pwdata', ApbDataType(32), doc=Doc(title='APB Data', ...))
    StructItem('penable', ApbEnaType(), doc=Doc(title='APB Transfer Enable', ...))
    StructItem('psel', ApbSelType(), doc=Doc(title='APB Slave Select', ...))
    StructItem('prdata', ApbDataType(32), orientation=BWD, doc=Doc(title='APB Data', ...))
    StructItem('pslverr', ApbRespType(), orientation=BWD, doc=Doc(title='APB Response Error', ...))
    StructItem('pready', ApbReadyType(), orientation=BWD, doc=Doc(title='APB Transfer Done', ...))

    With `pauser`:

    >>> class AuserType(ASecIdType):
    ...     def _build(self):
    ...         self._add(0, "apps")
    ...         self._add(2, "comm")
    ...         self._add(5, "audio")
    >>> apb5 = AmbaProto("APB5", ausertype=AuserType(default=2))

    >>> for item in ApbSlvType(proto=apb5).values(): item
    StructItem('paddr', ApbAddrType(12), doc=Doc(title='APB Bus Address', ...))
    StructItem('pauser', AuserType(default=2), doc=Doc(title='APB Address User Channel', ...))
    StructItem('pwrite', ApbWriteType(), doc=Doc(title='APB Write Enable', ...))
    StructItem('pwdata', ApbDataType(32), doc=Doc(title='APB Data', ...))
    StructItem('penable', ApbEnaType(), doc=Doc(title='APB Transfer Enable', ...))
    StructItem('psel', ApbSelType(), doc=Doc(title='APB Slave Select', ...))
    StructItem('prdata', ApbDataType(32), orientation=BWD, doc=Doc(title='APB Data', ...))
    StructItem('pslverr', ApbRespType(), orientation=BWD, doc=Doc(title='APB Response Error', ...))
    StructItem('pready', ApbReadyType(), orientation=BWD, doc=Doc(title='APB Transfer Done', ...))

    Both protocol versions are not connectable:

    >>> ApbSlvType().is_connectable(ApbSlvType(proto=apb5))
    False

    But casting is allowed:

    >>> for item in ApbSlvType().cast(ApbSlvType(proto=apb5)): item
    ('', '')
    ('paddr', 'paddr')
    ('pwrite', 'pwrite')
    ('pwdata', 'pwdata')
    ('penable', 'penable')
    ('psel', 'psel')
    ('prdata', 'prdata')
    ('pslverr', 'pslverr')
    ('pready', 'pready')
    """

    title: str = "APB Slave"
    comment: str = "APB Slave"
    proto: AmbaProto = AMBA3
    addrwidth: int = 12
    datawidth: int = 32

    def _build(self):
        # FWD
        self._add("paddr", ApbAddrType(self.addrwidth))
        if self.proto.ausertype:
            auserdoc = "APB Address User Channel"
            self._add("pauser", self.proto.ausertype, title=auserdoc, comment=auserdoc)
        if self.proto.wusertype:
            wuserdoc = "APB Write Data User Channel"
            self._add("pwuser", self.proto.wusertype, title=wuserdoc, comment=wuserdoc)
        self._add("pwrite", ApbWriteType())
        self._add("pwdata", ApbDataType(self.datawidth))
        if self.proto.has_wstrb:
            self._add("pstrb", ApbPstrbType(self.datawidth))
        self._add("penable", ApbEnaType())
        self._add("psel", ApbSelType())
        # BWD
        self._add("prdata", ApbDataType(self.datawidth), u.BWD)
        if rusertype := self.proto.rusertype:
            ruserdoc = "APB Read Data User Channel"
            self._add("pruser", rusertype, title=ruserdoc, comment=ruserdoc, dir=u.BWD)
        if busertype := self.proto.busertype:
            buserdoc = "APB Read Response User Channel"
            self._add("pbuser", busertype, title=buserdoc, comment=buserdoc, dir=u.BWD)
        self._add("pslverr", ApbRespType(), u.BWD)
        self._add("pready", ApbReadyType(), u.BWD)

    def cast(self, other):
        """
        How to cast an assign of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, ApbSlvType) and self.proto != other.proto:
            # Drive a Slv with Slv signals
            yield "", ""
            yield "paddr", "paddr"
            # never use pauser
            yield "pwrite", "pwrite"
            yield "pwdata", "pwdata"
            yield "penable", "penable"
            yield "psel", "psel"
            # BWD
            yield "prdata", "prdata"
            yield "pslverr", "pslverr"
            yield "pready", "pready"
        else:
            return None


class ApbSelType(u.BitType):
    """APB Select."""

    title: str = "APB Slave Select"
    comment: str = "APB Slave Select"


class ApbEnaType(u.EnaType):
    """APB Enable."""

    title: str = "APB Transfer Enable"
    comment: str = "APB Transfer Enable"


class ApbAddrType(u.UintType):
    """
    APB Address.

    >>> ApbAddrType().width
    32
    >>> ApbAddrType(width=16).width
    16

    Checked for valid width:
    >>> t = ApbAddrType(width=34)
    Traceback (most recent call last):
    ...
    ValueError: Illegal value for APB addrwidth: 34.Legal values are between 0 and 32.
    """

    title: str = "APB Bus Address"
    comment: str = "APB Bus Address"

    def __init__(self, width=32, **kwargs):
        if width not in LEGAL_APB_ADDR_WIDTH:
            raise ValueError(
                f"Illegal value for APB addrwidth: {width}."
                f"Legal values are between {LEGAL_APB_ADDR_WIDTH[0]} and {LEGAL_APB_ADDR_WIDTH[-1]}."
            )
        super().__init__(width=width, **kwargs)


class ApbDataType(u.UintType):
    """
    APB Data.

    >>> ApbDataType().width
    32
    >>> ApbDataType(width=16).width
    16

    Data width is checked for legal values:

    >>> t = ApbDataType(width=18)
    Traceback (most recent call last):
    ...
    ValueError: Illegal value for APB datawidth: 18. Legal values are [8, 16, 32]
    """

    title: str = "APB Data"
    comment: str = "APB Data"

    def __init__(self, width=32, **kwargs):
        if width not in LEGAL_APB_DATA_WIDTH:
            raise ValueError(f"Illegal value for APB datawidth: {width}. Legal values are {LEGAL_APB_DATA_WIDTH}")
        super().__init__(width=width, **kwargs)


class ApbPstrbType(u.UintType):
    """
    APB Write Strobe.

    Write strobe width is derived directly from data width.
    >>> ApbPstrbType().width
    4

    Data width is checked for legal values:
    >>> t = ApbPstrbType(29)
    Traceback (most recent call last):
    ...
    ValueError: Illegal value for APB datawidth: 29. Legal values are [8, 16, 32]
    """

    title: str = "APB Write Strobe"
    comment: str = "APB Write Strobe"

    def __init__(self, datawidth=32, **kwargs):
        if datawidth not in LEGAL_APB_DATA_WIDTH:
            raise ValueError(f"Illegal value for APB datawidth: {datawidth}. Legal values are {LEGAL_APB_DATA_WIDTH}")
        super().__init__(width=datawidth // 8, **kwargs)


class ApbProtType(u.UintType):
    """
    APB Protection Type.

    >>> ApbProtType().width
    3

    >>> ApbProtType().default
    0
    """

    title: str = "APB Protection Type"
    comment: str = "APB Protection Type"

    def __init__(self):
        super().__init__(width=3, default=0)


class ApbWriteType(u.AEnumType):
    """
    AHB Write Type.

    >>> for item in ApbWriteType().values(): item
    EnumItem(0, 'read', doc=Doc(title='Read operation'))
    EnumItem(1, 'write', doc=Doc(title='Write operation'))
    """

    keytype: u.BitType = u.BitType()
    title: str = "APB Write Enable"
    comment: str = "APB Write Enable"

    def _build(self):
        self._add(0, "read", "Read operation")
        self._add(1, "write", "Write operation")


class ApbRespType(u.AEnumType):
    """
    APB Response Type.

    >>> for item in ApbRespType().values(): item
    EnumItem(0, 'okay', doc=Doc(title='OK'))
    EnumItem(1, 'error', doc=Doc(title='Error'))
    """

    keytype: u.BitType = u.BitType()
    title: str = "APB Response Error"
    comment: str = "APB Response Error"

    def _build(self):
        self._add(0, "okay", "OK")
        self._add(1, "error", "Error")


class ApbReadyType(u.AEnumType):
    """
    APB Ready Type.

    >>> for item in ApbReadyType().values(): item
    EnumItem(0, 'busy', doc=Doc(title='Ongoing'))
    EnumItem(1, 'done', doc=Doc(title='Done'))
    """

    keytype: u.BitType = u.BitType(default=1)
    title: str = "APB Transfer Done"
    comment: str = "APB Transfer Done"

    def _build(self):
        self._add(0, "busy", "Ongoing")
        self._add(1, "done", "Done")


# class IdleType(u.AEnumType):
#     """
#     Bus Idle Type.

#     >>> for item in IdleType().values(): item
#     EnumItem(0, 'busy', doc=Doc(title='Busy', comment='Transfers Ongoing'))
#     EnumItem(1, 'idle', doc=Doc(title='Idle'))
#     """

#     keytype: u.BitType = u.BitType(default=1)
#     title: str = "Bus Idle"
#     comment: str = "Bus Idle"

#     def _build(self):
#         self._add(0, "busy", "Busy", comment="Transfers Ongoing")
#         self._add(1, "idle", "Idle")


def check_ahb_proto_pair(src_name: str, src_proto: AmbaProto, tgt_name: str, tgt_proto: AmbaProto) -> int:  # noqa: C901, PLR0912
    """
    Check AHB Protocol Compatibility.

    >>> p0 = AmbaProto("p0")
    >>> p1 = AmbaProto("p1",
    ...                 hmaster_width=3,
    ...                 hprotwidth=0,
    ...                 has_hburst=False,
    ...                 has_hmastlock=False,
    ...                 has_hnonsec=False,
    ...                 has_exclxfers=False
    ...               )
    >>> p2 = AmbaProto("p2",
    ...                hmaster_width=4,
    ...                hprotwidth=7,
    ...                has_hburst=True,
    ...                has_hmastlock=True,
    ...                has_hnonsec=True,
    ...                has_exclxfers=True
    ...               )
    >>> p3 = AmbaProto("p3", ausertype=u.UintType(3))
    >>> p4 = AmbaProto("p4", ausertype=u.UintType(4))
    >>> check_ahb_proto_pair("src", p0, "tgt", p0)
    0
    >>> check_ahb_proto_pair("src", p0, "tgt", p1)
    1
    >>> check_ahb_proto_pair("src", p1, "tgt", p2)
    1
    >>> check_ahb_proto_pair("src", p2, "tgt", p1)
    1
    >>> check_ahb_proto_pair("src", p0, "tgt", p3)
    1
    >>> check_ahb_proto_pair("src", p3, "tgt", p0)
    1
    >>> check_ahb_proto_pair("src", p3, "tgt", p4)
    2
    """
    if src_proto == tgt_proto:  # no need to check
        return 0

    opts = {
        "has_hburst": "hburst",
        "has_hmastlock": "hmastlock",
        "has_hnonsec": "hnonsec",
        "has_exclxfers": "hexcl/hexokay",
        "has_wstrb": "hwstrb",
    }
    src_dict = dict(src_proto)
    tgt_dict = dict(tgt_proto)
    verdict = 0
    pre1 = f"Protocol Pair for Source {src_name} with Protocol {src_proto.name} "
    pre2 = f"and Target {tgt_name} with Protocol {tgt_proto.name}"
    preamble = pre1 + pre2
    LOGGER.info(f"Checking {preamble}.")

    for usrtp in ["auser", "wuser", "ruser", "buser"]:
        tname = f"{usrtp}type"
        src_usertp = src_proto.__dict__[tname]
        tgt_usertp = tgt_proto.__dict__[tname]
        if src_usertp is not None and tgt_usertp is not None and src_usertp != tgt_usertp:
            LOGGER.error(f"{preamble}: Incompatible Definitions for 'h{usrtp}'!")
            verdict = 2

    if src_proto.ausertype is not None and tgt_proto.ausertype is None:
        LOGGER.warning(f"{preamble}: Ignoring source 'hauser'.")
        verdict = max(verdict, 1)
    elif src_proto.ausertype is None and tgt_proto.ausertype is not None:
        LOGGER.warning(f"{preamble}: Clamping target 'hauser'.")
        verdict = max(verdict, 1)
    if src_proto.hprotwidth > 0 and tgt_proto.hprotwidth == 0:
        LOGGER.warning(f"{preamble}: Ignoring source 'hprot'.")
        verdict = max(verdict, 1)
    elif src_proto.hprotwidth == 0 and tgt_proto.hprotwidth > 0:
        LOGGER.warning(f"{preamble}: Clamping target 'hprot'.")
        verdict = max(verdict, 1)
    elif src_proto.hprotwidth != tgt_proto.hprotwidth:
        LOGGER.warning(f"{preamble}: Different width for 'hprot' signals.")
        verdict = max(verdict, 1)
    if src_proto.hmaster_width > tgt_proto.hmaster_width:
        slc = "MSBs of " if tgt_proto.hmaster_width else ""
        LOGGER.warning(f"{preamble}: Ignoring {slc}source 'hmaster'.")
        verdict = max(verdict, 1)
    elif src_proto.hmaster_width < tgt_proto.hmaster_width:
        slc = "MSBs of " if src_proto.hmaster_width else ""
        LOGGER.warning(f"{preamble}: Clamping {slc}target 'hmaster'.")
        verdict = max(verdict, 1)

    for opt, sig in opts.items():
        if src_dict[opt] and not tgt_dict[opt]:
            LOGGER.warning(f"{preamble}: Ignoring source '{sig}'.")
            verdict = max(verdict, 1)
        elif not src_dict[opt] and tgt_dict[opt]:
            LOGGER.warning(f"{preamble}: Clamping target '{sig}'.")
            verdict = max(verdict, 1)

    return verdict



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - AHB2APB Tests.
"""

import logging
import random

import cocotb
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge
from humannum import Hex

from tests.ahb_driver import AHBMasterDriver, BurstType, SizeType
from tests.apb_driver import APBSlaveDriver


# TODO put this is a generic tb lib
async def wait_clocks(clock, cycles):
    """Helper Function."""
    for _ in range(cycles):
        await RisingEdge(clock)


@cocotb.test()
async def ahb2apb_test(dut):  # noqa: C901, PLR0912
    """Main Test Loop."""
    log = logging.getLogger(__name__)
    log.setLevel(logging.INFO)

    hclk = dut.main_clk_i
    rst_an = dut.main_rst_an_i

    ahb_mst = AHBMasterDriver(
        name="ahb_mst",
        log_level=logging.INFO,
        clk=hclk,
        rst_an=rst_an,
        hsel=dut.ahb_slv_hsel_i,
        haddr=dut.ahb_slv_haddr_i,
        hwrite=dut.ahb_slv_hwrite_i,
        hwstrb=dut.ahb_slv_hwstrb_i,
        hwdata=dut.ahb_slv_hwdata_i,
        htrans=dut.ahb_slv_htrans_i,
        hburst=dut.ahb_slv_hburst_i,
        hsize=dut.ahb_slv_hsize_i,
        hprot=dut.ahb_slv_hprot_i,
        hrdata=dut.ahb_slv_hrdata_o,
        hready=dut.ahb_slv_hreadyout_o,  # special hready/hreadyout swap as the master is not on a ML
        hreadyout=dut.ahb_slv_hready_i,
        hresp=dut.ahb_slv_hresp_o,
    )

    foo_slv = APBSlaveDriver(
        name="slv_foo",
        log_level=logging.INFO,
        clk=hclk,
        rst_an=rst_an,
        paddr=dut.apb_slv_foo_paddr_o,
        pwrite=dut.apb_slv_foo_pwrite_o,
        pwdata=dut.apb_slv_foo_pwdata_o,
        penable=dut.apb_slv_foo_penable_o,
        psel=dut.apb_slv_foo_psel_o,
        prdata=dut.apb_slv_foo_prdata_i,
        pready=dut.apb_slv_foo_pready_i,
        pslverr=dut.apb_slv_foo_pslverr_i,
        pready_delay=0,
        size_bytes=4 * 1024,
        err_addr={"w": list(range(0x400, 0x410)), "r": list(range(0x420, 0x430)), "rw": list(range(0x440, 0x450))},
    )

    bar_slv = APBSlaveDriver(
        name="slv_bar",
        log_level=logging.INFO,
        clk=hclk,
        rst_an=rst_an,
        paddr=dut.apb_slv_bar_paddr_o,
        pwrite=dut.apb_slv_bar_pwrite_o,
        pstrb=dut.apb_slv_bar_pstrb_o,
        pwdata=dut.apb_slv_bar_pwdata_o,
        penable=dut.apb_slv_bar_penable_o,
        psel=dut.apb_slv_bar_psel_o,
        prdata=dut.apb_slv_bar_prdata_i,
        pready=dut.apb_slv_bar_pready_i,
        pslverr=dut.apb_slv_bar_pslverr_i,
        pready_delay=0,
        size_bytes=1024,
    )

    baz_slv = APBSlaveDriver(
        name="slv_baz",
        log_level=logging.INFO,
        clk=hclk,
        rst_an=rst_an,
        paddr=dut.apb_slv_baz_paddr_o,
        pwrite=dut.apb_slv_baz_pwrite_o,
        pwdata=dut.apb_slv_baz_pwdata_o,
        penable=dut.apb_slv_baz_penable_o,
        psel=dut.apb_slv_baz_psel_o,
        prdata=dut.apb_slv_baz_prdata_i,
        pready=dut.apb_slv_baz_pready_i,
        pslverr=dut.apb_slv_baz_pslverr_i,
        pready_delay=0,
        size_bytes=13 * 1024,
    )

    mem = [bytearray(4 * 1024), bytearray(1024), bytearray(13 * 1024)]

    baseaddr = [0x0, 0x1000, 0x4000]

    btypes = (
        BurstType.SINGLE,
        BurstType.WRAP4,
        BurstType.INCR4,
        BurstType.WRAP8,
        BurstType.INCR8,
        BurstType.WRAP16,
        BurstType.INCR16,
    )

    sizes = (SizeType.BYTE, SizeType.HALFWORD, SizeType.WORD)

    slaves = [foo_slv, bar_slv, baz_slv]

    cocotb.start_soon(Clock(hclk, period=10).start())

    cocotb.start_soon(foo_slv.run())
    cocotb.start_soon(bar_slv.run())
    cocotb.start_soon(baz_slv.run())

    # initial reset
    rst_an.value = 0
    await wait_clocks(hclk, 10)
    rst_an.value = 1
    await wait_clocks(hclk, 10)

    # randomized accesses
    for _ in range(50):
        tgt = 1  # random.randint(0, 2)
        btype = random.choice(btypes)
        if slaves[tgt].pstrb is not None:
            size = random.choice(sizes)
        else:
            size = SizeType.WORD
        if btype == BurstType.SINGLE:
            blen = 1
            mmask = (1 << size) - 1
        else:
            blen = 2 << (btype >> 1)
            mmask = (4 << (((btype - 2) >> 1) + size)) - 1
        offs = random.randint(0, 100) & (~((1 << size) - 1))  # make it size aligned
        if btype in (BurstType.INCR16, BurstType.INCR8, BurstType.INCR4):
            offs &= ~mmask  # make it burst aligned

        smax = (1 << (1 << (size + 3))) - 1  # max value according to size
        if random.randint(0, 1):
            wdata = [random.randint(1, smax) for i in range(blen)]
            refdata = ahb_mst.calc_wrmem(offs=offs, size=size, blen=blen, mmask=mmask, wdata=wdata, mem=mem[tgt])
            mem[tgt][(offs & ~mmask) : (offs & ~mmask) + (blen << size)] = refdata
            log.info(
                f"=MST WRITE TRANSFER= target: {slaves[tgt].name}; offs:{hex(offs)}; "
                f"burst:{btype.name}; size:{size.name}; "
                f"wdata:{[hex(w) for w in wdata]}\n"
            )
            err_resp = await ahb_mst.write(baseaddr[tgt] + offs, wdata, burst_type=btype, size=size)
            assert not err_resp, "Unexpected error response"
            for ln in range(5):
                log.warning(f"MEM[{ln*16}:{ln*16+15}] = {[str(Hex(w, 8)) for w in mem[tgt][(ln*16):(ln*16+15)]]}")
        else:
            xdata = ahb_mst.calc_expected(offs=offs, size=size, blen=blen, mmask=mmask, mem=mem[tgt])
            err_resp, rdata = await ahb_mst.read(baseaddr[tgt] + offs, burst_type=btype, size=size)
            assert not err_resp, "Unexpected error response"
            if tuple(rdata) == tuple(xdata):
                log.info(
                    f"=MST READ TRANSFER= target: {slaves[tgt].name}; offs:{hex(offs)}; "
                    f"burst:{btype.name}; size:{size.name};\n"
                    f"> rdata: {[hex(w) for w in rdata]};"
                )
            else:
                log.error(
                    f"=MST READ TRANSFER MISMATCH= target: {slaves[tgt].name}; offs:{hex(offs)}; "
                    f"burst:{btype.name}; size:{size.name};\n"
                    f"> expected: {[hex(w) for w in xdata]};\n"
                    f"> got:      {[hex(w) for w in rdata]};"
                )
                raise AssertionError("Read data compare mismatch.")

        await wait_clocks(hclk, random.randint(0, 5))
    await wait_clocks(hclk, 20)

    # directed tests for expected error responses
    # 0x400...0x40F error on write
    # 0x420...0x42F error on read
    # 0x440...0x44F error in read and write

    log.info("=MST WRITE TRANSFER= target: 0; offs:0x408; with expected Error Response")
    err_resp = await ahb_mst.write(0x0000408, 0xDEAD)
    assert err_resp, "Missed expected Error Response"
    await wait_clocks(hclk, 5)

    err_resp, rdata = await ahb_mst.read(0x0000408)
    assert not err_resp, "Unexpected error response"
    if tuple(rdata) == (0,):
        log.info(f"=MST READ TRANSFER= target: 0; offs:0x408;\n> rdata: {[hex(w) for w in rdata]};")
    else:
        log.error(
            "=MST READ TRANSFER MISMATCH= target: 0; offs:0x408;\n"
            "> expected: [0x0];\n"
            f"> got:      {[hex(w) for w in rdata]};"
        )
        raise AssertionError("Read data compare mismatch.")
    await wait_clocks(hclk, 5)

    log.info("=MST WRITE TRANSFER= target: 0; offs:0x424; wdata:[0xAFFE]")
    err_resp = await ahb_mst.write(0x0000424, 0xAFFE)
    assert not err_resp, "Unexpected error response"
    await wait_clocks(hclk, 5)

    log.info("=MST READ TRANSFER= target: 0; offs:0x424; with expected Error Response")
    err_resp, rdata = await ahb_mst.read(0x0000424)
    assert err_resp, "Missed expected Error Response"
    if tuple(rdata) != (0,):
        log.error(
            "=MST READ TRANSFER MISMATCH= target: 0; offs:0x424;\n"
            "> expected: [0x0];\n"
            f"> got:      {[hex(w) for w in rdata]};"
        )
        raise AssertionError("Read data compare mismatch.")
    await wait_clocks(hclk, 5)

    log.info("=MST WRITE TRANSFER= target: 0; offs:0x44C; with expected Error Response")
    err_resp = await ahb_mst.write(0x000044C, 0xDEAD)
    assert err_resp, "Missed expected Error Response"
    await wait_clocks(hclk, 5)

    log.info("=MST READ TRANSFER= target: 0; offs:0x44C; with expected Error Response")
    err_resp, rdata = await ahb_mst.read(0x000044C)
    assert err_resp, "Missed expected Error Response"
    if tuple(rdata) != (0,):
        log.error(
            "=MST READ TRANSFER MISMATCH= target: 0; offs:0x44C;\n"
            "> expected: [0x0];\n"
            f"> got:      {[hex(w) for w in rdata]};"
        )
        raise AssertionError("Read data compare mismatch.")

    await wait_clocks(hclk, 30)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - APB2MEM.
"""

from typing import ClassVar

import ucdp as u
from ucdp_glbl.mem import MemIoType

from . import types as t


class UcdpApb2memMod(u.ATailoredMod):
    """APB to MEMio Converter."""

    datawidth: int = 32
    """Data Width in Bits."""
    addrwidth: int = 12
    """Address Width in Bits."""
    proto: t.AmbaProto = t.AmbaProto()
    """AMBA Protocol Specifier."""

    filelists: ClassVar[u.ModFileLists] = (
        u.ModFileList(
            name="hdl",
            gen="full",
            filepaths=("$PRJROOT/{mod.topmodname}/{mod.modname}.sv"),
            template_filepaths=("ucdp_apb2mem.sv.mako", "sv.mako"),
        ),
    )

    def _build(self):
        self.add_port(
            t.ApbSlvType(proto=self.proto, addrwidth=self.addrwidth, datawidth=self.datawidth),
            "apb_slv_i",
            title="APB Slave Input",
        )
        self.add_port(
            MemIoType(addrwidth=self.addrwidth - 2, datawidth=self.datawidth, writable=True, err=True),
            "mem_o",
            title="Memory Interface",
        )

    @staticmethod
    def build_top(**kwargs):
        """Build example top module and return it."""
        return UcdpApb2MemExampleMod()


class UcdpApb2MemExampleMod(u.AMod):
    """Example Converter."""

    def _build(self):
        UcdpApb2memMod(self, "u_a2m", datawidth=16, addrwidth=10)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - AHB Tests.
"""

import logging
import random

import cocotb
from cocotb.clock import Clock
from cocotb.triggers import Combine, RisingEdge

from tests.ahb_driver import AHBMasterDriver, AHBSlaveDriver, BurstType, SizeType


# TODO put this is a generic tb lib
async def wait_clocks(clock, cycles):
    """Helper Function."""
    for _ in range(cycles):
        await RisingEdge(clock)


@cocotb.test()
async def ahb_ml_test(dut):
    """Main Test Loop."""
    log = logging.getLogger(__name__)
    log.setLevel(logging.INFO)

    hclk = dut.main_clk_i
    rst_an = dut.main_rst_an_i

    ext_mst = AHBMasterDriver(
        name="ext_mst",
        log_level=logging.INFO,
        clk=hclk,
        rst_an=rst_an,
        haddr=dut.ahb_mst_ext_haddr_i,
        hauser=dut.ahb_mst_ext_hauser_i,
        hmaster=dut.ahb_mst_ext_hmaster_i,
        hwrite=dut.ahb_mst_ext_hwrite_i,
        hwdata=dut.ahb_mst_ext_hwdata_i,
        hwstrb=dut.ahb_mst_ext_hwstrb_i,
        htrans=dut.ahb_mst_ext_htrans_i,
        hburst=dut.ahb_mst_ext_hburst_i,
        hsize=dut.ahb_mst_ext_hsize_i,
        hprot=dut.ahb_mst_ext_hprot_i,
        hrdata=dut.ahb_mst_ext_hrdata_o,
        hready=dut.ahb_mst_ext_hready_o,
        hresp=dut.ahb_mst_ext_hresp_o,
        hsel=None,
    )

    dsp_mst = AHBMasterDriver(
        name="dsp_mst",
        log_level=logging.INFO,
        clk=hclk,
        rst_an=rst_an,
        haddr=dut.ahb_mst_dsp_haddr_i,
        hmaster=dut.ahb_mst_dsp_hmaster_i,
        hwrite=dut.ahb_mst_dsp_hwrite_i,
        hwdata=dut.ahb_mst_dsp_hwdata_i,
        hwstrb=dut.ahb_mst_dsp_hwstrb_i,
        htrans=dut.ahb_mst_dsp_htrans_i,
        hburst=dut.ahb_mst_dsp_hburst_i,
        hsize=dut.ahb_mst_dsp_hsize_i,
        hprot=dut.ahb_mst_dsp_hprot_i,
        hrdata=dut.ahb_mst_dsp_hrdata_o,
        hready=dut.ahb_mst_dsp_hready_o,
        hresp=dut.ahb_mst_dsp_hresp_o,
        hsel=None,
    )

    ram_slv = AHBSlaveDriver(
        name="ram_slv",
        log_level=logging.INFO,
        clk=hclk,
        rst_an=rst_an,
        hsel=dut.ahb_slv_ram_hsel_o,
        haddr=dut.ahb_slv_ram_haddr_o,
        hauser=dut.ahb_slv_ram_hauser_o,
        hmaster=dut.ahb_slv_ram_hmaster_o,
        hwrite=dut.ahb_slv_ram_hwrite_o,
        hwstrb=dut.ahb_slv_ram_hwstrb_o,
        htrans=dut.ahb_slv_ram_htrans_o,
        hsize=dut.ahb_slv_ram_hsize_o,
        hburst=dut.ahb_slv_ram_hburst_o,
        hprot=dut.ahb_slv_ram_hprot_o,
        hwdata=dut.ahb_slv_ram_hwdata_o,
        hready=dut.ahb_slv_ram_hready_o,
        hreadyout=dut.ahb_slv_ram_hreadyout_i,
        hresp=dut.ahb_slv_ram_hresp_i,
        hrdata=dut.ahb_slv_ram_hrdata_i,
    )

    cocotb.start_soon(Clock(hclk, period=10).start())

    cocotb.start_soon(ram_slv.run())

    hauser_width = len(ext_mst.hauser) if ext_mst.hauser else 0
    hmaster_width = len(ext_mst.hmaster) if ext_mst.hmaster else 0

    # initial reset
    rst_an.value = 0
    await wait_clocks(hclk, 10)
    rst_an.value = 1
    await wait_clocks(hclk, 10)

    await dsp_mst.write(0xF0000300, 0xBEEFBEEF)
    await wait_clocks(hclk, 5)

    ext_wr = cocotb.start_soon(ext_mst.write(0xF0000300, 0x76543210))
    dsp_wr = cocotb.start_soon(
        dsp_mst.write(
            0xF0000316, (0x11, 0x22, 0x33, 0x44), burst_type=BurstType.WRAP4, size=SizeType.HALFWORD, hmaster=5
        )
    )
    await Combine(ext_wr, dsp_wr)
    await wait_clocks(hclk, 5)
    # ram_slv.log_data()

    # ext_wr = cocotb.start_soon(ext_mst.write(0xF0000000, 0x76543210))
    # dsp_wr = cocotb.start_soon(
    #     dsp_mst.write(0xF0000016, (0x11, 0x22, 0x33, 0x44), burst_type=BurstType.WRAP4, size=SizeType.HALFWORD)
    # )
    # await Combine(ext_wr, dsp_wr)

    # await wait_clocks(hclk, 5)
    # rdata = await ext_mst.read(0xF0000000, burst_type=BurstType.INCR8, size=SizeType.BYTE)
    # print("MST EXT rdata:", [hex(data) for data in rdata])
    # await wait_clocks(hclk, 5)

    mem = bytearray(1024)
    btypes = (
        BurstType.SINGLE,
        BurstType.WRAP4,
        BurstType.INCR4,
        BurstType.WRAP8,
        BurstType.INCR8,
        BurstType.WRAP16,
        BurstType.INCR16,
    )
    sizes = (SizeType.BYTE, SizeType.HALFWORD, SizeType.WORD)
    for _ in range(20):
        btype = random.choice(btypes)
        size = random.choice(sizes)
        if btype == BurstType.SINGLE:
            blen = 1
            mmask = (1 << size) - 1
        else:
            blen = 2 << (btype >> 1)
            mmask = (4 << (((btype - 2) >> 1) + size)) - 1
        offs = random.randint(0, 511) & ~((1 << size) - 1)  # make it size aligned

        if btype in (BurstType.INCR16, BurstType.INCR8, BurstType.INCR4):
            offs &= ~mmask  # make it burst aligned
        valmax = (1 << (1 << (size + 3))) - 1  # max value according to size
        strbmax = (1 << (1 << size)) - 1
        hauser = random.randint(1, (1 << hauser_width) - 1)
        hmaster = random.randint(1, (1 << hmaster_width) - 1)
        if random.randint(0, 1):
            wdata = [random.randint(1, valmax) for i in range(blen)]
            wstrb = [random.randint(1, strbmax) for i in range(blen)]

            mem[(offs & ~mmask) : (offs & ~mmask) + (blen << size)] = ext_mst.calc_wrmem(
                offs=offs, size=size, blen=blen, mmask=mmask, wdata=wdata, wstrb=wstrb, mem=mem
            )
            log.info(
                f"=MST WRITE TRANSFER= offs:{hex(offs)}; burst:{btype.name}; size:{size.name}; "
                f"wdata:{[hex(w) for w in wdata]}; wstrb:{[hex(w) for w in wstrb]}"
            )
            err_resp = await ext_mst.write(
                0xF0000000 + offs, wdata, wstrb=wstrb, burst_type=btype, size=size, hauser=hauser, hmaster=hmaster
            )
            assert not err_resp, "Unexpected error response"
        else:
            xdata = ext_mst.calc_expected(offs=offs, size=size, blen=blen, mmask=mmask, mem=mem)
            err_resp, rdata = await ext_mst.read(
                0xF0000000 + offs, burst_type=btype, size=size, hauser=hauser, hmaster=hmaster
            )
            assert not err_resp, "Unexpected error response"
            if tuple(rdata) == tuple(xdata):
                log.info(
                    f"=MST READ TRANSFER= offs:{hex(offs)}; burst:{btype.name}; size:{size.name};\n"
                    f"> rdata: {[hex(w) for w in rdata]};"
                )
            else:
                log.error(
                    f"=MST READ TRANSFER MISMATCH= offs:{hex(offs)}; burst:{btype.name}; size:{size.name};\n"
                    f"> expected: {[hex(w) for w in xdata]};\n"
                    f"> got:      {[hex(w) for w in rdata]};"
                )
                raise AssertionError("Read data compare mismatch.")

        await wait_clocks(hclk, 2)

    await wait_clocks(hclk, 30)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - AHB Bus Converter.
"""

import enum
from logging import getLogger
from typing import Any, ClassVar, NamedTuple

import ucdp as u
from aligntext import Align, center, left
from icdutil import num

from . import types as t

LOGGER = getLogger(__name__)


class ExtSpec(enum.Enum):
    """Extension Specifier."""

    PORT = 1
    AUTO = 2


PORT = ExtSpec.PORT
AUTO = ExtSpec.AUTO


class ConvDescr(NamedTuple):
    """Conversion Specifier."""

    src_pres: bool = True
    tgt_pres: bool = True
    conv: str = "forward"


OPTMAP = [["n/a", "tie-off"], ["ignore", "forward"]]
DIRMAP = [["m2m", "m2s"], ["s2m", "s2s"]]
VECMAP = [["wide to narrow", "narrow to wide"], ["narrow to wide", "wide to narrow"]]
UTPMAP = [["u2u", "u2e"], ["e2u", "e2e"]]


class UcdpAhb2ahbMod(u.ATailoredMod):
    """
    AHB-to-AHB Converter.

    Universal AHB Bus Converter:
        Master to Slave,
        Slave to Master,
        Optional Protocol Conversion,
        Data and Address Width Conversion,
        Clock Domain Crossing
    """

    filelists: ClassVar[u.ModFileLists] = (
        u.ModFileList(
            name="hdl",
            gen="full",
            filepaths=("$PRJROOT/{mod.topmodname}/{mod.modname}.sv"),
            template_filepaths=("ucdp_ahb2ahb.sv.mako", "sv.mako"),
        ),
    )

    source_type: t.AhbMstType | t.AhbSlvType
    target_type: t.AhbMstType | t.AhbSlvType
    async_conv: bool = False
    _conv_spec: dict[str, Any] = u.PrivateField(default_factory=dict)
    __ext_param: dict[str, Any] = u.PrivateField(default_factory=dict)

    def _get_usertype_width(self, usertpye) -> int:
        if usertpye is None:
            return 0
        if isinstance(usertpye, u.UintType):
            return usertpye.width
        if isinstance(usertpye, u.AEnumType):
            return usertpye.bits
        LOGGER.error(f"Dealing with unknown user type: {usertpye}")
        return 0

    def _determine_haddr_conv(self) -> None:
        src_type = self.source_type
        tgt_type = self.target_type
        awdiff = tgt_type.addrwidth - src_type.addrwidth
        self._conv_spec["haddr"] = {}
        conv = "forward" if awdiff == 0 else "reduce" if awdiff < 0 else "expand"
        if self.__ext_param.get("haddr", None) is None:
            if conv == "reduce":
                self._conv_spec["haddr"]["red"] = tgt_type.addrwidth
            elif conv == "expand":
                self._conv_spec["haddr"]["exp"] = u.UintType(awdiff, default=0)
                conv += " by constant"
        elif conv == "expand":
            self._conv_spec["haddr"]["exp"] = self.__ext_param["haddr"]
            if isinstance(self.__ext_param["haddr"], u.Port):
                conv += " by port"
            else:
                conv += " by constant"
        self._conv_spec["haddr"]["descr"] = ConvDescr(conv=conv)

    def _determine_hdata_hstrb_conv(self) -> None:
        src_type = self.source_type
        tgt_type = self.target_type
        # hwdata/hrdata
        self._conv_spec["hwdata"] = {}
        self._conv_spec["hrdata"] = {}
        dwdiff = tgt_type.datawidth - src_type.datawidth
        if dwdiff == 0:
            conv = "forward"
        else:
            conv = "convert"
            if dwdiff > 0:
                wide, narr = tgt_type.datawidth, src_type.datawidth
            else:
                wide, narr = src_type.datawidth, tgt_type.datawidth
            self._conv_spec["hwdata"]["narr"] = narr
            self._conv_spec["hwdata"]["ratio"] = wide // narr
            self._conv_spec["hwdata"]["n2w"] = dwdiff > 0
        self._conv_spec["hwdata"]["descr"] = ConvDescr(conv=conv + (f" {VECMAP[0][dwdiff > 0]}" if dwdiff != 0 else ""))
        self._conv_spec["hrdata"]["descr"] = ConvDescr(conv=conv + (f" {VECMAP[1][dwdiff > 0]}" if dwdiff != 0 else ""))
        # hwstrb
        self._conv_spec["hwstrb"] = {}
        strbw = tgt_type.datawidth >> 3
        conv = OPTMAP[src_type.proto.has_wstrb][tgt_type.proto.has_wstrb]
        if conv == "tie-off":
            self._conv_spec["hwstrb"]["tie"] = u.UintType(strbw, default=(1 << strbw) - 1)
        elif (conv == "forward") and (dwdiff != 0):
            conv = "convert" + (f" {VECMAP[0][dwdiff > 0]}" if dwdiff != 0 else "")
            self._conv_spec["hwstrb"]["narr"] = narr >> 3
            self._conv_spec["hwstrb"]["ratio"] = wide // narr
            self._conv_spec["hwstrb"]["n2w"] = dwdiff > 0
        self._conv_spec["hwstrb"]["descr"] = ConvDescr(
            src_pres=src_type.proto.has_wstrb, tgt_pres=tgt_type.proto.has_wstrb, conv=conv
        )

    def _determine_usertp_conv(self, usrsig: str) -> None:  # noqa: C901, PLR0912
        self._conv_spec[usrsig] = {}
        tname = usrsig[1:] + "type"  # drop the 'h'
        src_usertp = self.source_type.proto.__dict__[tname]
        tgt_usertp = self.target_type.proto.__dict__[tname]
        if usrsig in ["hruser", "hbuser"]:
            src_usertp, tgt_usertp = tgt_usertp, src_usertp  # swap direction
        conv = OPTMAP[src_usertp is not None][tgt_usertp is not None]
        utp = UTPMAP[isinstance(src_usertp, u.AEnumType)][isinstance(tgt_usertp, u.AEnumType)]
        self._conv_spec[usrsig]["utp"] = utp
        usertp_ext = self.__ext_param.get(usrsig, None)
        if conv == "tie-off":
            if usertp_ext is None or usertp_ext == AUTO:
                if isinstance(tgt_usertp, u.AEnumType):
                    self._conv_spec[usrsig]["tie"] = tgt_usertp.decode(tgt_usertp.default)
                else:
                    self._conv_spec[usrsig]["tie"] = tgt_usertp
            else:
                self._conv_spec[usrsig]["tie"] = self.__ext_param[usrsig]
        elif conv == "forward":
            if usertp_ext is None:
                if src_usertp != tgt_usertp:
                    LOGGER.error(f"Mismatching '{usrsig}' types without explicit conversion given.")
                    conv = "tie-off"
                    self._conv_spec[usrsig]["tie"] = tgt_usertp
            else:
                src_hw = self._get_usertype_width(src_usertp)
                tgt_hw = self._get_usertype_width(tgt_usertp)
                if isinstance(usertp_ext, dict):
                    conv = "convert"
                    self._conv_spec[usrsig]["map"] = usertp_ext
                elif tgt_hw < src_hw:
                    conv = "reduce"
                    self._conv_spec[usrsig]["red"] = tgt_hw
                elif tgt_hw > src_hw:
                    if isinstance(usertp_ext, u.Port):
                        conv = "expand by port"
                        self._conv_spec[usrsig]["exp"] = usertp_ext
                    else:
                        conv = "expand by constant"
                        if usertp_ext == AUTO:
                            self._conv_spec[usrsig]["exp"] = u.UintType(tgt_hw - src_hw, default=0)
                        else:
                            self._conv_spec[usrsig]["exp"] = usertp_ext
        self._conv_spec[usrsig]["descr"] = ConvDescr(
            src_pres=src_usertp is not None, tgt_pres=tgt_usertp is not None, conv=conv
        )

    def _determine_hmaster_conv(self) -> None:
        src_type = self.source_type
        tgt_type = self.target_type
        self._conv_spec["hmaster"] = {}
        src_hmw = src_type.proto.hmaster_width
        tgt_hmw = tgt_type.proto.hmaster_width
        hmwdiff = tgt_hmw - src_hmw
        conv = OPTMAP[src_hmw > 0][tgt_hmw > 0]
        if self.__ext_param.get("hmaster", None) is None:
            if conv == "tie-off":
                self._conv_spec["hmaster"]["tie"] = u.UintType(tgt_hmw, default=0)
            elif conv == "forward" and hmwdiff < 0:
                conv = "reduce"
                self._conv_spec["hmaster"]["red"] = tgt_hmw
            elif conv == "forward" and hmwdiff > 0:
                conv = "expand by constant"
                self._conv_spec["hmaster"]["exp"] = u.UintType(hmwdiff, default=0)
        elif hmwdiff > 0:
            if conv == "tie-off":
                self._conv_spec["hmaster"]["tie"] = self.__ext_param["hmaster"]
            else:
                conv = "expand by port" if isinstance(self.__ext_param["hmaster"], u.Port) else "expand by constant"
                self._conv_spec["hmaster"]["exp"] = self.__ext_param["hmaster"]
        self._conv_spec["hmaster"]["descr"] = ConvDescr(src_pres=src_hmw > 0, tgt_pres=tgt_hmw > 0, conv=conv)

    def _determine_hprot_conv(self) -> None:
        src_type = self.source_type
        tgt_type = self.target_type
        self._conv_spec["hprot"] = {}
        src_hpw = src_type.proto.hprotwidth
        tgt_hpw = tgt_type.proto.hprotwidth
        hpwdiff = tgt_hpw - src_hpw
        conv = OPTMAP[src_hpw > 0][tgt_hpw > 0]
        hprot_ext = self.__ext_param.get("hprot", None)
        if conv == "forward":
            if hpwdiff > 0:
                conv = "expand"
                self._conv_spec["hprot"]["exp"] = hprot_ext
                if hprot_ext is None:
                    conv += " by logic"
                elif isinstance(hprot_ext, u.Port):
                    conv += " by port"
                else:
                    conv += " by constant"
            elif hpwdiff < 0:
                conv = "reduce"
        elif conv == "tie-off":
            self._conv_spec["hprot"]["tie"] = hprot_ext
            if hprot_ext is None:
                self._conv_spec["hprot"]["tie"] = tgt_type.proto.hprottype
                conv += " to default"
            elif isinstance(hprot_ext, u.Port):
                conv += " to port"
            else:
                conv += " to constant"
        self._conv_spec["hprot"]["descr"] = ConvDescr(src_pres=src_hpw > 0, tgt_pres=tgt_hpw > 0, conv=conv)

    def _determine_conv(self):
        if self._conv_spec != {}:
            return
        src_type = self.source_type
        tgt_type = self.target_type
        src_slv = isinstance(src_type, t.AhbSlvType)
        tgt_slv = isinstance(tgt_type, t.AhbSlvType)
        self._conv_spec["dir"] = convdir = DIRMAP[src_slv][tgt_slv]
        # simple forwarding:
        self._conv_spec["hwrite"] = {"descr": ConvDescr()}
        self._conv_spec["htrans"] = {"descr": ConvDescr()}
        self._conv_spec["hsize"] = {"descr": ConvDescr()}
        self._conv_spec["hresp"] = {"descr": ConvDescr()}
        # hsel
        self._conv_spec["hsel"] = {
            "descr": ConvDescr(src_pres=src_slv, tgt_pres=tgt_slv, conv=OPTMAP[src_slv][tgt_slv])
        }
        # haddr
        self._determine_haddr_conv()
        # hwdata/hrdata/hwstrb
        self._determine_hdata_hstrb_conv()
        # hburst/hnonsec/hmastlock
        for opt in ["hburst", "hnonsec", "hmastlock"]:
            src_pres = src_type.proto.__dict__[f"has_{opt}"]
            tgt_pres = tgt_type.proto.__dict__[f"has_{opt}"]
            self._conv_spec[opt] = {
                "descr": ConvDescr(src_pres=src_pres, tgt_pres=tgt_pres, conv=OPTMAP[src_pres][tgt_pres])
            }
        # hauser/hwuser/hruser/hbuser
        self._determine_usertp_conv("hauser")
        self._determine_usertp_conv("hwuser")
        self._determine_usertp_conv("hruser")
        self._determine_usertp_conv("hbuser")
        # hmaster
        self._determine_hmaster_conv()
        # hexcl/hexokay
        src_excl = src_type.proto.has_exclxfers
        tgt_excl = tgt_type.proto.has_exclxfers
        self._conv_spec["hexcl"] = {
            "descr": ConvDescr(src_pres=src_excl, tgt_pres=tgt_excl, conv=OPTMAP[src_excl][tgt_excl])
        }
        # deliberately swapped optmap
        self._conv_spec["hexokay"] = {
            "descr": ConvDescr(src_pres=src_excl, tgt_pres=tgt_excl, conv=OPTMAP[tgt_excl][src_excl])
        }
        # hready/hreadyout
        conv = "forward" if convdir in ["m2m", "s2s"] else "convert"
        self._conv_spec["hready"] = {"descr": ConvDescr(conv=conv)}
        conv = "forward" if convdir == "s2s" else "n/a" if convdir == "m2m" else "convert"
        self._conv_spec["hreadyout"] = {"descr": ConvDescr(src_pres=src_slv, tgt_pres=tgt_slv, conv=conv)}
        # hprot
        self._determine_hprot_conv()

    def _add_usertp_const(self) -> None:
        for usrsig in ["hauser", "hwuser", "hruser", "hbuser"]:
            if self._conv_spec[usrsig]["descr"].conv in ["tie-off", "convert"]:
                tname = usrsig[1:] + "type"  # drop the 'h'
                utp = self._conv_spec[usrsig]["utp"]
                # hruser & hbuser are BWD signals
                srcd = ["e2u", "e2e"] if usrsig in ["hauser", "hwuser"] else ["u2e", "e2e"]
                tgtd = ["u2e", "e2e"] if usrsig in ["hauser", "hwuser"] else ["e2u", "e2e"]
                if utp in srcd:
                    src_usertp = self.source_type.proto.__dict__[tname]
                    self.add_type_consts(src_usertp, name=f"src_{usrsig}")
                if utp in tgtd:
                    tgt_usertp = self.target_type.proto.__dict__[tname]
                    self.add_type_consts(tgt_usertp, name=f"tgt_{usrsig}")

    def _build(self):
        if self.async_conv:
            self.add_port(u.ClkRstAnType(), "src_i")
            self.add_port(u.ClkRstAnType(), "tgt_i")
        elif self.source_type.datawidth != self.target_type.datawidth:
            self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(self.source_type, "ahb_src_i", title="AHB Source Input", comment="AHB Source")
        self.add_port(self.target_type, "ahb_tgt_o", title="AHB Target Output", comment="AHB Target")

    def _build_dep(self):
        # TODO: async
        if self.source_type != self.target_type:
            self.add_type_consts(t.AhbTransType())
            self.add_type_consts(t.AhbBurstType())
        self._determine_conv()
        self._add_usertp_const()

    def define_haddr_ext(self, ext: int | ExtSpec = 0) -> None:
        """Define extension for haddr when src_addrwith < tgt_addrwidth."""
        if self.__ext_param.get("haddr", None) is not None:
            LOGGER.warning("Redefining HADDR extension")
        awdiff = self.target_type.addrwidth - self.source_type.addrwidth
        if awdiff <= 0:
            LOGGER.warning(f"Source addrwidth equal or greater than target addrwidth, no extension for {self.name}.")
            return
        if isinstance(ext, int):
            if num.calc_unsigned_width(ext) > awdiff:
                LOGGER.error(f"Extension value of {ext} too wide for haddr extension.")
                ext = 0
            self.__ext_param["haddr"] = u.UintType(awdiff, default=ext)
        elif ext == PORT:
            port = self.add_port(
                u.UintType(awdiff), "haddr_ext_i", title="Target HADDR Extension", comment="Target HADDR Extension"
            )
            self.__ext_param["haddr"] = port

    def define_hmaster_ext(self, ext: int | ExtSpec = 0) -> None:
        """Define extension of hmaster when width of src.hmaster_width < tgt.hmaster_width (including tie-off)."""
        if self.__ext_param.get("hmaster", None) is not None:
            LOGGER.warning("Redefining HMASTER extension")
        hmwdiff = self.target_type.proto.hmaster_width - self.source_type.proto.hmaster_width
        if hmwdiff <= 0:
            LOGGER.warning(
                f"Source hmaster_width equal or greater than target hmaster_width, no extension for {self.name}."
            )
            return
        if isinstance(ext, int):
            if num.calc_unsigned_width(ext) > hmwdiff:
                LOGGER.error(f"Extension value of {ext} too wide for hmaster extension.")
                ext = 0
            self.__ext_param["hmaster"] = u.UintType(hmwdiff, default=ext)
        elif ext == PORT:
            port = self.add_port(
                u.UintType(hmwdiff),
                "hmaster_ext_i",
                title="Target HMASTER Extension",
                comment="Target HMASTER Extension",
            )
            self.__ext_param["hmaster"] = port

    def define_hprot_ext(self, ext: int | ExtSpec = 0) -> None:
        """Define extension of hprot when width of src.hprotwidth < tgt.hprotwidth (including tie-off)."""
        if self.__ext_param.get("hprot", None) is not None:
            LOGGER.warning("Redefining HPROT extension")
        hpwdiff = self.target_type.proto.hprotwidth - self.source_type.proto.hprotwidth
        if hpwdiff <= 0:
            LOGGER.warning(f"Source hprotwidth equal or greater than target hprotwidth, no extension for {self.name}.")
            return
        if isinstance(ext, int):
            if num.calc_unsigned_width(ext) > hpwdiff:
                LOGGER.error(f"Extension value of {ext} too wide for hprot extension.")
                ext = 0
            self.__ext_param["hprot"] = u.UintType(hpwdiff, default=ext)
        elif ext == PORT:
            port = self.add_port(
                u.UintType(hpwdiff),
                "hprot_ext_i",
                title="Target HPROT Extension",
                comment="Target HPROT Extension",
            )
            self.__ext_param["hprot"] = port

    def _user_conversion(self, usrsig: str, conv: int | str | dict | ExtSpec = 0) -> None:  # noqa: C901, PLR0912
        """Prepare conversion for user-types."""
        uname = usrsig.upper()
        if self.__ext_param.get(usrsig, None) is not None:
            LOGGER.warning(f"Redefining {uname} conversion")
        tname = usrsig[1:] + "type"  # drop the 'h'
        src_usertp = self.source_type.proto.__dict__[tname]
        tgt_usertp = self.target_type.proto.__dict__[tname]
        if usrsig in ["hruser", "hbuser"]:
            src_usertp, tgt_usertp = tgt_usertp, src_usertp  # swap direction
            snd, rcv = "tgt", "src"
        else:
            snd, rcv = "src", "tgt"
        if (src_usertp is None or tgt_usertp is None) and isinstance(conv, dict):
            LOGGER.error(f"Conversion map given, but not both interfaces have {uname}.")
            return
        if src_usertp is not None and src_usertp == tgt_usertp and not isinstance(conv, dict):
            LOGGER.warning("{uname} types are identical, no conversion necessary.")
            return
        if conv == AUTO:
            self.__ext_param[usrsig] = AUTO
            return
        src_uw = self._get_usertype_width(src_usertp)
        tgt_uw = self._get_usertype_width(tgt_usertp)
        uwdiff = tgt_uw - src_uw
        if isinstance(conv, int):
            if uwdiff <= 0:
                LOGGER.warning(
                    f"Source {uname} width equal or greater than {rcv}_{uname} width, no extension for {self.name}."
                )
                return
            if num.calc_unsigned_width(conv) > uwdiff:
                LOGGER.error(f"Extension value of {conv} too wide for {uname} extension.")
                conv = 0
            if uwdiff == tgt_uw and isinstance(tgt_usertp, u.AEnumType):  # tie-off for enum
                if conv not in tgt_usertp.keys():
                    raise ValueError(f"Value '{conv}' can not be used to index tgt_{usrsig}")
                self.__ext_param[usrsig] = tgt_usertp.decode(conv)
            else:
                self.__ext_param[usrsig] = u.UintType(uwdiff, default=conv)
        elif isinstance(conv, str):
            if not isinstance(tgt_usertp, u.AEnumType):
                raise ValueError(f"Value '{conv}' does not match {uname} type.")
            if uwdiff != tgt_uw:
                raise ValueError(f"Value '{conv}' can only be used for tie-off, not for extension.")
            if conv not in [item.value for item in tgt_usertp.values()]:
                raise ValueError(f"Value '{conv}' is not in {uname} enumeration.")
            self.__ext_param[usrsig] = conv
        elif conv == PORT:
            if uwdiff <= 0:
                LOGGER.warning(
                    f"Source {uname} width equal or greater than {rcv}_{uname} width, no extension for {self.name}."
                )
                return
            port = self.add_port(
                u.UintType(uwdiff),
                f"{usrsig}_ext_i",
                title=f"Target {uname} Extension",
                comment=f"Target {uname} Extension",
            )
            self.__ext_param[usrsig] = port
        elif isinstance(conv, dict):
            if isinstance(src_usertp, u.AEnumType):
                src_values = [item.value for item in src_usertp.values()]
            if isinstance(tgt_usertp, u.AEnumType):
                tgt_values = [item.value for item in tgt_usertp.values()]
            map = {}
            for key, value in conv.items():
                kk, vv = key, value
                if key is None:
                    map[kk] = vv
                    continue
                if isinstance(key, int):
                    if isinstance(src_usertp, u.UintType):
                        if num.calc_unsigned_width(key) > src_uw:
                            raise ValueError(f"Value '{key}' too big for {snd}_{usrsig} vector.")
                        kk = u.UintType(src_uw, default=key)
                    if isinstance(src_usertp, u.AEnumType):
                        if key not in src_usertp.keys():
                            raise ValueError(f"Value '{key}' is not in keys of {snd}_{usrsig}.")
                        kk = src_usertp.decode(key)
                    map[kk] = vv
                    continue
                if isinstance(key, str):
                    if not isinstance(src_usertp, u.AEnumType):
                        raise ValueError(f"Value '{key}' can not be used to index {snd}_{usrsig}.")
                    if key not in src_values:
                        raise ValueError(f"Value '{key}' is not in {snd}_{usrsig} enumeration.")
                    map[kk] = vv
            for key, value in map.items():
                kk, vv = key, value
                if isinstance(value, int):
                    if isinstance(tgt_usertp, u.UintType):
                        if num.calc_unsigned_width(value) > tgt_uw:
                            raise ValueError(f"Value '{value}' too big for {rcv}_{usrsig} vector.")
                        vv = u.UintType(tgt_uw, default=value)
                    if isinstance(tgt_usertp, u.AEnumType):
                        if value not in tgt_usertp.keys():
                            raise ValueError(f"Value '{value}' can not be used to index {rcv}_{usrsig}")
                        vv = tgt_usertp.decode(value)
                    map[kk] = vv
                    continue
                if isinstance(value, str):
                    if not isinstance(tgt_usertp, u.AEnumType):
                        raise ValueError(f"Value '{value}' can not be used to index {rcv}_{usrsig}.")
                    if value not in tgt_values:
                        raise ValueError(f"Value '{value}' is not in {rcv}_{usrsig} enumeration.")
                    map[kk] = vv
                    continue
                raise ValueError(f"Value '{value}' can not be mapped to {rcv}_{usrsig}")
            if isinstance(tgt_usertp, u.AEnumType):
                map.setdefault(None, tgt_usertp.decode(tgt_usertp.default))
            else:
                map.setdefault(None, u.UintType(tgt_uw, default=tgt_usertp.default))
            self.__ext_param[usrsig] = map

    def define_hauser_conversion(self, conv: int | str | dict | ExtSpec = 0) -> None:
        """Define conversion for 'hauser' types."""
        self._user_conversion("hauser", conv)

    def define_hwuser_conversion(self, conv: int | str | dict | ExtSpec = 0) -> None:
        """Define conversion for 'hwuser' types."""
        self._user_conversion("hwuser", conv)

    def define_hruser_conversion(self, conv: int | str | dict | ExtSpec = 0) -> None:
        """Define conversion for 'hruser' types."""
        self._user_conversion("hruser", conv)

    def define_hbuser_conversion(self, conv: int | str | dict | ExtSpec = 0) -> None:
        """Define conversion for 'hbuser' types."""
        self._user_conversion("hbuser", conv)

    def get_overview(self) -> str:
        """Converter Overview."""
        if self.source_type == self.target_type:
            return "No conversion, just feed-through."
        self._determine_conv()
        cdir = self._conv_spec["dir"].replace("s", "Slave")
        cdir = cdir.replace("m", "Master")
        cdir = cdir.replace("2", " to ")
        ovr = Align(rtrim=True)
        ovr.set_alignments(left, center, center, left)
        ovr.add_spacer("Converting " + cdir)
        ovr.add_spacer()
        ovr.add_row("Signal", "Src", "Tgt", "Conv")
        for signal, item in self._conv_spec.items():
            if not isinstance(item, dict) or (conv := item.get("descr", None)) is None:
                continue
            src_pres = "x" if conv.src_pres else "-"
            tgt_pres = "x" if conv.tgt_pres else "-"
            ovr.add_row(signal, src_pres, tgt_pres, conv.conv)
        return "\n".join(ovr)

    @staticmethod
    def build_top(**kwargs):
        """Build example top module and return it."""
        return UcdpAhb2ahbExampleMod()


class UcdpAhb2ahbExampleMod(u.AMod):
    """
    Just an Example Converter.
    """

    def _build(self):  # noqa: C901, PLR0912
        class MyUserType(t.ASecIdType):
            """My AUser Type."""

            title: str = "AHB User Type"
            comment: str = "AHB User Type"

            def _build(self):
                self._add(0, "apps")
                self._add(2, "comm")
                self._add(5, "audio")

        class MyOtherUserType(t.ASecIdType):
            """My Other AUser Type."""

            title: str = "Another AHB User Type"
            comment: str = "Another AHB User Type"
            keytype: u.UintType = u.UintType(6)

            def _build(self):
                self._add(0, "audio")
                self._add(2, "apps")
                self._add(5, "comm")

        minp = t.AmbaProto(
            name="minp",
            hprotwidth=0,
            has_hburst=False,
            has_wstrb=False,
            hmaster_width=0,
            has_hmastlock=False,
            has_hnonsec=False,
            has_exclxfers=False,
        )

        smlp = t.AmbaProto(
            name="smlp",
            has_hburst=True,
            has_wstrb=True,
            hmaster_width=4,
            has_hmastlock=True,
            has_hnonsec=True,
            has_exclxfers=True,
        )

        lrgp = t.AmbaProto(
            name="lrgp",
            hprotwidth=7,
            has_wstrb=True,
            hmaster_width=6,
            has_hmastlock=True,
            has_hnonsec=True,
            has_exclxfers=True,
        )

        hsp = minp.new(
            name="hsp",
            ausertype=u.UintType(5, default=1),
            wusertype=u.UintType(5, default=1),
            rusertype=u.UintType(5, default=1),
            busertype=u.UintType(5, default=1),
        )
        hep = minp.new(
            name="hep",
            ausertype=MyUserType(default=2),
            wusertype=MyUserType(default=2),
            rusertype=MyUserType(default=2),
            busertype=MyUserType(default=2),
        )
        hop = minp.new(
            name="hop",
            ausertype=MyOtherUserType(default=5),
            wusertype=MyOtherUserType(default=5),
            rusertype=MyOtherUserType(default=5),
            busertype=MyOtherUserType(default=5),
        )

        for srctp, srctt in [(t.AhbMstType, "mst"), (t.AhbSlvType, "slv")]:
            for tgttp, tgttt in [(t.AhbMstType, "mst"), (t.AhbSlvType, "slv")]:
                for srcp in [minp, smlp, lrgp]:
                    srcaw = 36 if srcp == lrgp else 32
                    srcdw = 128 if srcp == lrgp else 32
                    for tgtp in [minp, smlp, lrgp]:
                        tgtaw = 36 if tgtp == lrgp else 32
                        tgtdw = 128 if tgtp == lrgp else 32
                        source_type = srctp(addrwidth=srcaw, datawidth=srcdw, proto=srcp)
                        target_type = tgttp(addrwidth=tgtaw, datawidth=tgtdw, proto=tgtp)
                        for ext, extn in [(None, "n"), (PORT, "p"), (1, "s")]:
                            mname = f"u_{srctt}2{tgttt}_{srcp.name}_{tgtp.name}_{extn}"
                            m = UcdpAhb2ahbMod(
                                self,
                                mname,
                                source_type=source_type,
                                target_type=target_type,
                            )
                            if ext is None:
                                continue
                            m.define_haddr_ext(ext)
                            m.define_hmaster_ext(ext)
                            m.define_hprot_ext(ext)

        for srcp in [minp, hsp, hep, hop]:
            for tgtp in [minp, hsp, hep, hop]:
                if (srcp == minp) and (tgtp == minp):
                    continue  # combination already covered above
                for trsl, tname in [(None, "n"), (AUTO, "a"), (PORT, "p"), (2, "s"), ("apps", "e")]:
                    if tname == "n" and srcp in [hsp, hep, hop] and tgtp in [hsp, hep, hop] and srcp != tgtp:
                        continue  # different hauser types need explicit conversion
                    mname = f"u_usertp_{srcp.name}_{tgtp.name}_{tname}"
                    m = UcdpAhb2ahbMod(
                        self,
                        mname,
                        source_type=t.AhbMstType(proto=srcp),
                        target_type=t.AhbMstType(proto=tgtp),
                    )
                    if trsl is None:
                        continue
                    # when using enum value, it must be for tie-off and signal to be tied must be an enum
                    if tname != "e" or ((srcp == minp) and (tgtp in [hep, hop])):
                        m.define_hauser_conversion(trsl)
                        m.define_hwuser_conversion(trsl)
                    if tname != "e" or ((tgtp == minp) and (srcp in [hep, hop])):
                        m.define_hruser_conversion(trsl)
                        m.define_hbuser_conversion(trsl)

        hmap_u2u = {0: 5, 2: 0, 5: 2, None: 2}
        hmap_u2e = {0: "comm", 2: 0, 5: 2}
        hmap_e2u = {0: 5, "apps": 0, 5: 2}
        hmap_e2e = {0: "comm", "apps": 0, 5: 2, None: "audio"}

        for srcp in [hsp, hep, hop]:
            for tgtp in [hsp, hep, hop]:
                if srcp in [hep, hop] and srcp == tgtp:
                    continue  # no need to map identical enum, it's just forwarding
                for trsl, trbk, tname in [
                    (hmap_u2u, hmap_u2u, "u2u"),
                    (hmap_u2e, hmap_e2u, "u2e"),
                    (hmap_e2u, hmap_u2e, "e2u"),
                    (hmap_e2e, hmap_e2e, "e2e"),
                ]:
                    if (srcp == hsp and tname[0] == "e") or (tgtp == hsp and tname[2] == "e"):
                        continue  # can't use enum values for non-enum hauser
                    mname = f"u_usertp_{srcp.name}_{tgtp.name}_{tname}"
                    m = UcdpAhb2ahbMod(
                        self,
                        mname,
                        source_type=t.AhbMstType(proto=srcp),
                        target_type=t.AhbMstType(proto=tgtp),
                    )
                    m.define_hauser_conversion(trsl)
                    m.define_hwuser_conversion(trsl)
                    m.define_hruser_conversion(trbk)
                    m.define_hbuser_conversion(trbk)
